"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BinaryRxServer = void 0;
const messages_1 = require("./messages");
const rxjs_1 = require("rxjs");
const operators_1 = require("rxjs/operators");
const TimedQueue_1 = require("../json-rx/TimedQueue");
const codec_1 = require("./codec");
const util_1 = require("../json-rx/util");
const isUint8Array_1 = require("../util/isUint8Array");
const ERR_UNKNOWN = new Uint8Array([0]);
const ERR_ID_TAKEN = new Uint8Array([1]);
const ERR_TOO_MANY_SUBSCRIPTIONS = new Uint8Array([2]);
const ERR_INVALID_DATA = new Uint8Array([3]);
class BinaryRxServer {
    constructor({ send, call, notify, maxActiveSubscriptions = 30, bufferSize = 10, bufferTime = 1, }) {
        this.active = new Map();
        this.encoder = new codec_1.Encoder();
        this.call = call;
        this.notify = notify;
        this.maxActiveSubscriptions = maxActiveSubscriptions;
        if (bufferTime) {
            const buffer = new TimedQueue_1.TimedQueue();
            buffer.itemLimit = bufferSize;
            buffer.timeLimit = bufferTime;
            buffer.onFlush = (messages) => {
                send(this.encoder.encode(messages));
            };
            this.send = (message) => buffer.push(message);
        }
        else {
            this.send = (message) => {
                send(this.encoder.encode([message]));
            };
        }
    }
    sendError(id, data) {
        this.send(new messages_1.ErrorMessage(id, data));
    }
    onSubscribe(message, ctx) {
        const { id, method, data } = message;
        util_1.assertName(method);
        try {
            const active = this.active.get(id);
            if (active) {
                this.active.delete(id);
                active.unsubscribe();
                this.sendError(id, ERR_ID_TAKEN);
                return;
            }
            if (this.active.size >= this.maxActiveSubscriptions)
                return this.sendError(id, ERR_TOO_MANY_SUBSCRIPTIONS);
            const callResult = this.call(method, data, ctx);
            const observable = (rxjs_1.isObservable(callResult)
                ? callResult
                : rxjs_1.from(callResult).pipe(operators_1.mergeMap((value) => (rxjs_1.isObservable(value) ? value : rxjs_1.of(value)))));
            const ref = { buffer: [] };
            let done = false;
            const subscription = observable.subscribe((data) => {
                if (!isUint8Array_1.isUint8Array(data))
                    return this.sendError(id, ERR_INVALID_DATA);
                if (data === undefined)
                    return;
                ref.buffer.push(data);
                util_1.microtask(() => {
                    if (!ref.buffer.length)
                        return;
                    try {
                        for (const payload of ref.buffer)
                            this.send(new messages_1.DataMessage(id, payload));
                    }
                    finally {
                        ref.buffer = [];
                    }
                });
            }, (error) => {
                done = true;
                this.active.delete(id);
                this.sendError(id, isUint8Array_1.isUint8Array(error) ? error : ERR_UNKNOWN);
            }, () => {
                done = true;
                this.active.delete(id);
                try {
                    if (!ref.buffer.length)
                        return this.send(new messages_1.CompleteMessage(id, undefined));
                    const last = ref.buffer.length - 1;
                    for (let i = 0; i <= last; i++) {
                        const isLast = i === last;
                        const message = isLast
                            ? new messages_1.CompleteMessage(id, ref.buffer[i])
                            : new messages_1.DataMessage(id, ref.buffer[i]);
                        this.send(message);
                    }
                }
                finally {
                    ref.buffer = [];
                }
            });
            if (!done)
                this.active.set(id, subscription);
        }
        catch (error) {
            this.sendError(id, Buffer.isBuffer(error) ? error : ERR_UNKNOWN);
        }
    }
    onUnsubscribe(message) {
        const { id } = message;
        const subscription = this.active.get(id);
        if (!subscription)
            return;
        this.active.delete(id);
        subscription.unsubscribe();
    }
    onNotification(message, ctx) {
        const { method, data } = message;
        util_1.assertName(method);
        this.notify(method, data, ctx);
    }
    onArray(arr, ctx) {
        const messages = codec_1.decodeFullMessages(arr, 0);
        this.onMessages(messages, ctx);
    }
    onMessages(messages, ctx) {
        const length = messages.length;
        for (let i = 0; i < length; i++) {
            const message = messages[i];
            if (message instanceof messages_1.NotificationMessage)
                this.onNotification(message, ctx);
            else if (message instanceof messages_1.SubscribeMessage)
                this.onSubscribe(message, ctx);
            else if (message instanceof messages_1.UnsubscribeMessage)
                this.onUnsubscribe(message);
        }
    }
    stop() {
        this.send = (message) => { };
        for (const sub of this.active.values()) {
            sub.unsubscribe();
        }
    }
}
exports.BinaryRxServer = BinaryRxServer;
