"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PatchBuilder = void 0;
const clock_1 = require("./clock");
const DeleteOperation_1 = require("./operations/DeleteOperation");
const InsertArrayElementsOperation_1 = require("./operations/InsertArrayElementsOperation");
const InsertStringSubstringOperation_1 = require("./operations/InsertStringSubstringOperation");
const MakeArrayOperation_1 = require("./operations/MakeArrayOperation");
const MakeNumberOperation_1 = require("./operations/MakeNumberOperation");
const MakeObjectOperation_1 = require("./operations/MakeObjectOperation");
const MakeStringOperation_1 = require("./operations/MakeStringOperation");
const SetObjectKeysOperation_1 = require("./operations/SetObjectKeysOperation");
const SetRootOperation_1 = require("./operations/SetRootOperation");
const SetNumberOperation_1 = require("./operations/SetNumberOperation");
const NoopOperation_1 = require("./operations/NoopOperation");
const MakeConstantOperation_1 = require("./operations/MakeConstantOperation");
const MakeValueOperation_1 = require("./operations/MakeValueOperation");
const SetValueOperation_1 = require("./operations/SetValueOperation");
const constants_1 = require("./constants");
const Patch_1 = require("./Patch");
class PatchBuilder {
    constructor(clock) {
        this.clock = clock;
        this.patch = new Patch_1.Patch();
    }
    obj() {
        this.pad();
        const id = this.clock.tick(1);
        const op = new MakeObjectOperation_1.MakeObjectOperation(id);
        this.patch.ops.push(op);
        return id;
    }
    arr() {
        this.pad();
        const id = this.clock.tick(1);
        const op = new MakeArrayOperation_1.MakeArrayOperation(id);
        this.patch.ops.push(op);
        return id;
    }
    str() {
        this.pad();
        const id = this.clock.tick(1);
        const op = new MakeStringOperation_1.MakeStringOperation(id);
        this.patch.ops.push(op);
        return id;
    }
    num() {
        this.pad();
        const id = this.clock.tick(1);
        const op = new MakeNumberOperation_1.MakeNumberOperation(id);
        this.patch.ops.push(op);
        return id;
    }
    const(value) {
        this.pad();
        const id = this.clock.tick(1);
        const op = new MakeConstantOperation_1.MakeConstantOperation(id, value);
        this.patch.ops.push(op);
        return id;
    }
    val(value) {
        this.pad();
        const id = this.clock.tick(1);
        const op = new MakeValueOperation_1.MakeValueOperation(id, value);
        this.patch.ops.push(op);
        return id;
    }
    root(value) {
        this.pad();
        const id = this.clock.tick(1);
        const op = new SetRootOperation_1.SetRootOperation(id, value);
        this.patch.ops.push(op);
        return id;
    }
    setKeys(obj, tuples) {
        this.pad();
        if (!tuples.length)
            throw new Error('EMPTY_TUPLES');
        const id = this.clock.tick(1);
        const op = new SetObjectKeysOperation_1.SetObjectKeysOperation(id, obj, tuples);
        const span = op.span();
        if (span > 1)
            this.clock.tick(span - 1);
        this.patch.ops.push(op);
        return id;
    }
    setNum(obj, value) {
        this.pad();
        const id = this.clock.tick(1);
        const op = new SetNumberOperation_1.SetNumberOperation(id, obj, value);
        this.patch.ops.push(op);
        return id;
    }
    setVal(obj, value) {
        this.pad();
        const id = this.clock.tick(1);
        const op = new SetValueOperation_1.SetValueOperation(id, obj, value);
        this.patch.ops.push(op);
        return id;
    }
    insStr(obj, after, substring) {
        this.pad();
        if (!substring.length)
            throw new Error('EMPTY_STRING');
        const id = this.clock.tick(1);
        const op = new InsertStringSubstringOperation_1.InsertStringSubstringOperation(id, obj, after, substring);
        const span = op.span();
        if (span > 1)
            this.clock.tick(span - 1);
        this.patch.ops.push(op);
        return id;
    }
    insArr(arr, after, elements) {
        this.pad();
        const id = this.clock.tick(1);
        const op = new InsertArrayElementsOperation_1.InsertArrayElementsOperation(id, arr, after, elements);
        const span = op.span();
        if (span > 1)
            this.clock.tick(span - 1);
        this.patch.ops.push(op);
        return id;
    }
    del(obj, start, span) {
        this.pad();
        const id = this.clock.tick(span);
        const op = new DeleteOperation_1.DeleteOperation(id, obj, start, span);
        this.patch.ops.push(op);
        return id;
    }
    noop(span) {
        this.pad();
        const id = this.clock.tick(span);
        const op = new NoopOperation_1.NoopOperation(id, span);
        this.patch.ops.push(op);
        return id;
    }
    jsonObj(json) {
        const obj = this.obj();
        const keys = Object.keys(json);
        if (keys.length) {
            const tuples = [];
            for (const k of keys)
                tuples.push([k, this.json(json[k])]);
            this.setKeys(obj, tuples);
        }
        return obj;
    }
    jsonArr(json) {
        const arr = this.arr();
        if (json.length) {
            const values = [];
            for (const el of json)
                values.push(this.json(el));
            this.insArr(arr, arr, values);
        }
        return arr;
    }
    jsonStr(json) {
        const str = this.str();
        if (json)
            this.insStr(str, str, json);
        return str;
    }
    jsonVal(json) {
        return this.val(json);
    }
    json(json) {
        switch (json) {
            case null:
                return constants_1.NULL_ID;
            case true:
                return constants_1.TRUE_ID;
            case false:
                return constants_1.FALSE_ID;
        }
        if (Array.isArray(json))
            return this.jsonArr(json);
        switch (typeof json) {
            case 'object':
                return this.jsonObj(json);
            case 'string':
                return this.jsonStr(json);
            case 'number':
                return this.jsonVal(json);
        }
        return constants_1.UNDEFINED_ID;
    }
    pad() {
        const nextTime = this.patch.nextTime();
        if (!nextTime)
            return;
        const drift = this.clock.time - nextTime;
        if (drift > 0) {
            const id = new clock_1.LogicalTimestamp(this.clock.getSessionId(), nextTime);
            const padding = new NoopOperation_1.NoopOperation(id, drift);
            this.patch.ops.push(padding);
        }
    }
}
exports.PatchBuilder = PatchBuilder;
