"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.encode = exports.encodeTimestamp = void 0;
const encodeString_1 = require("../../../util/encodeString");
const DeleteOperation_1 = require("../../operations/DeleteOperation");
const InsertArrayElementsOperation_1 = require("../../operations/InsertArrayElementsOperation");
const InsertStringSubstringOperation_1 = require("../../operations/InsertStringSubstringOperation");
const MakeArrayOperation_1 = require("../../operations/MakeArrayOperation");
const MakeNumberOperation_1 = require("../../operations/MakeNumberOperation");
const MakeObjectOperation_1 = require("../../operations/MakeObjectOperation");
const MakeStringOperation_1 = require("../../operations/MakeStringOperation");
const NoopOperation_1 = require("../../operations/NoopOperation");
const SetNumberOperation_1 = require("../../operations/SetNumberOperation");
const SetObjectKeysOperation_1 = require("../../operations/SetObjectKeysOperation");
const SetRootOperation_1 = require("../../operations/SetRootOperation");
const varuint_1 = require("./util/varuint");
const encodeTimestamp = (ts) => {
    const sessionId = ts.getSessionId();
    let low32 = sessionId | 0;
    if (low32 < 0)
        low32 += 4294967296;
    const high8 = (sessionId - low32) / 4294967296;
    return [low32, ((high8 << 24) | ts.time) >>> 0];
};
exports.encodeTimestamp = encodeTimestamp;
const encode = (patch) => {
    const { ops } = patch;
    const buffers = [new Uint32Array(exports.encodeTimestamp(patch.getId())).buffer];
    let size = 8;
    for (const op of ops) {
        if (op instanceof MakeObjectOperation_1.MakeObjectOperation) {
            buffers.push(new Uint8Array([0]));
            size += 1;
            continue;
        }
        if (op instanceof MakeArrayOperation_1.MakeArrayOperation) {
            buffers.push(new Uint8Array([1]));
            size += 1;
            continue;
        }
        if (op instanceof MakeStringOperation_1.MakeStringOperation) {
            buffers.push(new Uint8Array([2]));
            size += 1;
            continue;
        }
        if (op instanceof MakeNumberOperation_1.MakeNumberOperation) {
            buffers.push(new Uint8Array([3]));
            size += 1;
            continue;
        }
        if (op instanceof SetRootOperation_1.SetRootOperation) {
            buffers.push(new Uint8Array([4]));
            buffers.push(new Uint32Array(exports.encodeTimestamp(op.value)).buffer);
            size += 1 + 8;
            continue;
        }
        if (op instanceof SetObjectKeysOperation_1.SetObjectKeysOperation) {
            buffers.push(new Uint8Array([5]));
            buffers.push(new Uint32Array(exports.encodeTimestamp(op.object)).buffer);
            size += 1 + 8;
            const keyNumberBuffer = new Uint8Array(varuint_1.encodeVarUInt(op.tuples.length));
            buffers.push(keyNumberBuffer);
            size += keyNumberBuffer.byteLength;
            for (const [key, value] of op.tuples) {
                const valueBuffer = new Uint32Array(exports.encodeTimestamp(value)).buffer;
                const keyBuffer = encodeString_1.encodeString(key);
                const keyLengthBuffer = new Uint8Array(varuint_1.encodeVarUInt(keyBuffer.byteLength));
                buffers.push(valueBuffer, keyLengthBuffer, keyBuffer);
                size += valueBuffer.byteLength + keyLengthBuffer.byteLength + keyBuffer.byteLength;
            }
            continue;
        }
        if (op instanceof SetNumberOperation_1.SetNumberOperation) {
            buffers.push(new Uint8Array([6]), new Uint32Array(exports.encodeTimestamp(op.num)).buffer, new Float64Array([op.value]).buffer);
            size += 1 + 8 + 8;
            continue;
        }
        if (op instanceof InsertStringSubstringOperation_1.InsertStringSubstringOperation) {
            const stringBuffer = encodeString_1.encodeString(op.substring);
            const stringLengthBuffer = new Uint8Array(varuint_1.encodeVarUInt(stringBuffer.byteLength));
            buffers.push(new Uint8Array([7]), new Uint32Array(exports.encodeTimestamp(op.obj)).buffer, new Uint32Array(exports.encodeTimestamp(op.after)).buffer, stringLengthBuffer.buffer, stringBuffer);
            size += 1 + 8 + 8 + stringLengthBuffer.byteLength + stringBuffer.byteLength;
            continue;
        }
        if (op instanceof InsertArrayElementsOperation_1.InsertArrayElementsOperation) {
            const { arr, after, elements } = op;
            const length = elements.length;
            const elementLengthBuffer = new Uint8Array(varuint_1.encodeVarUInt(length));
            buffers.push(new Uint8Array([8]), new Uint32Array(exports.encodeTimestamp(arr)).buffer, new Uint32Array(exports.encodeTimestamp(after)).buffer, elementLengthBuffer.buffer);
            for (const element of elements)
                buffers.push(new Uint32Array(exports.encodeTimestamp(element)).buffer);
            size += 1 + 8 + 8 + elementLengthBuffer.byteLength + 8 * length;
            continue;
        }
        if (op instanceof DeleteOperation_1.DeleteOperation) {
            const { obj, after, length } = op;
            if (length > 1) {
                const spanBuffer = new Uint8Array(varuint_1.encodeVarUInt(length));
                buffers.push(new Uint8Array([9]), new Uint32Array(exports.encodeTimestamp(obj)).buffer, new Uint32Array(exports.encodeTimestamp(after)).buffer, spanBuffer.buffer);
                size += 1 + 8 + 8 + spanBuffer.byteLength;
                continue;
            }
            buffers.push(new Uint8Array([10]), new Uint32Array(exports.encodeTimestamp(obj)).buffer, new Uint32Array(exports.encodeTimestamp(after)).buffer);
            size += 1 + 8 + 8;
            continue;
        }
        if (op instanceof NoopOperation_1.NoopOperation) {
            const { length } = op;
            if (length > 1) {
                const spanBuffer = new Uint8Array(varuint_1.encodeVarUInt(length));
                buffers.push(new Uint8Array([12]), spanBuffer.buffer);
                size += 1 + spanBuffer.byteLength;
                continue;
            }
            buffers.push(new Uint8Array([11]));
            size += 1;
            continue;
        }
    }
    const res = new Uint8Array(size);
    let offset = 0;
    for (const buffer of buffers) {
        res.set(new Uint8Array(buffer), offset);
        offset += buffer.byteLength;
    }
    return res;
};
exports.encode = encode;
