"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.AbstractEncoder = void 0;
const Encoder_1 = require("../../../json-pack/Encoder");
const utf8_1 = require("../../../util/utf8");
const ConstantType_1 = require("../../types/const/ConstantType");
const ObjectType_1 = require("../../types/lww-object/ObjectType");
const ValueType_1 = require("../../types/lww-value/ValueType");
const ArrayType_1 = require("../../types/rga-array/ArrayType");
const StringType_1 = require("../../types/rga-string/StringType");
class AbstractEncoder extends Encoder_1.Encoder {
    encodeRoot(root) {
        this.ts(root.id);
        if (root.node)
            this.encodeNode(root.node);
    }
    encodeNode(node) {
        if (node instanceof ObjectType_1.ObjectType)
            return this.encodeObj(node);
        else if (node instanceof ArrayType_1.ArrayType)
            return this.encodeArr(node);
        else if (node instanceof StringType_1.StringType)
            return this.encodeStr(node);
        else if (node instanceof ValueType_1.ValueType)
            return this.encodeVal(node);
        else if (node instanceof ConstantType_1.ConstantType)
            return this.encodeConst(node);
        throw new Error('UNKNOWN_NODE');
    }
    encodeObj(obj) {
        const length = obj.latest.size;
        this.encodeObjectHeader(length);
        this.ts(obj.id);
        for (const [key, chunk] of obj.latest.entries())
            this.encodeObjChunk(key, chunk);
    }
    encodeObjChunk(key, chunk) {
        this.ts(chunk.id);
        const length = utf8_1.utf8Count(key);
        this.vuint57(length);
        this.encodeUtf8(key, length);
        this.encodeNode(chunk.node);
    }
    encodeArr(obj) {
        const length = obj.size();
        this.encodeArrayHeader(length);
        this.ts(obj.id);
        for (const chunk of obj.chunks())
            this.encodeArrChunk(chunk);
    }
    encodeArrChunk(chunk) {
        if (chunk.deleted) {
            this.b1vuint56(true, chunk.deleted);
            this.ts(chunk.id);
        }
        else {
            const nodes = chunk.nodes;
            const length = nodes.length;
            this.b1vuint56(false, length);
            this.ts(chunk.id);
            for (let i = 0; i < length; i++)
                this.encodeNode(nodes[i]);
        }
    }
    encodeStr(obj) {
        const length = obj.size();
        this.encodeStringHeader(length);
        this.ts(obj.id);
        for (const chunk of obj.chunks())
            this.encodeStrChunk(chunk);
    }
    encodeStrChunk(chunk) {
        if (chunk.deleted) {
            this.b1vuint56(true, chunk.deleted);
            this.ts(chunk.id);
        }
        else {
            const text = chunk.str;
            const length = utf8_1.utf8Count(text);
            this.b1vuint56(false, length);
            this.ts(chunk.id);
            this.encodeUtf8(text, length);
        }
    }
    encodeConst(obj) {
        const { value } = obj;
        switch (value) {
            case null:
            case undefined:
            case false:
            case true:
                this.encodeAny(value);
                break;
            default: {
                if (typeof value === 'number')
                    this.encodeNumber(value);
                else {
                    this.u8(0xd4);
                    this.encodeAny(value);
                }
            }
        }
    }
    encodeVal(obj) {
        this.u8(0xd5);
        this.ts(obj.id);
        this.ts(obj.writeId);
        this.encodeAny(obj.value);
    }
    encodeUtf8(str, byteLength) {
        this.ensureCapacity(byteLength);
        const strLength = str.length;
        const uint8 = this.uint8;
        let offset = this.offset;
        let pos = 0;
        while (pos < strLength) {
            let value = str.charCodeAt(pos++);
            if ((value & 0xffffff80) === 0) {
                uint8[offset++] = value;
                continue;
            }
            else if ((value & 0xfffff800) === 0) {
                uint8[offset++] = ((value >> 6) & 0x1f) | 0xc0;
            }
            else {
                if (value >= 0xd800 && value <= 0xdbff) {
                    if (pos < strLength) {
                        const extra = str.charCodeAt(pos);
                        if ((extra & 0xfc00) === 0xdc00) {
                            pos++;
                            value = ((value & 0x3ff) << 10) + (extra & 0x3ff) + 0x10000;
                        }
                    }
                }
                if ((value & 0xffff0000) === 0) {
                    uint8[offset++] = ((value >> 12) & 0x0f) | 0xe0;
                    uint8[offset++] = ((value >> 6) & 0x3f) | 0x80;
                }
                else {
                    uint8[offset++] = ((value >> 18) & 0x07) | 0xf0;
                    uint8[offset++] = ((value >> 12) & 0x3f) | 0x80;
                    uint8[offset++] = ((value >> 6) & 0x3f) | 0x80;
                }
            }
            uint8[offset++] = (value & 0x3f) | 0x80;
        }
        this.offset = offset;
    }
    id(x, y) {
        if (x <= 0b111 && y <= 0b1111) {
            this.u8((x << 4) | y);
        }
        else {
            this.b1vuint28(true, x);
            this.vuint39(y);
        }
    }
    vuint57(num) {
        if (num <= 0b1111111) {
            this.u8(num);
        }
        else if (num <= 16383) {
            this.u8(0b10000000 | (num & 0b01111111));
            this.u8(num >>> 7);
        }
        else if (num <= 2097151) {
            this.u8(0b10000000 | (num & 0b01111111));
            this.u8(0b10000000 | ((num >>> 7) & 0b01111111));
            this.u8(num >>> 14);
        }
        else if (num <= 268435455) {
            this.u8(0b10000000 | (num & 0b01111111));
            this.u8(0b10000000 | ((num >>> 7) & 0b01111111));
            this.u8(0b10000000 | ((num >>> 14) & 0b01111111));
            this.u8(num >>> 21);
        }
        else {
            let lo32 = num | 0;
            if (lo32 < 0)
                lo32 += 4294967296;
            const hi32 = (num - lo32) / 4294967296;
            if (num <= 34359738367) {
                this.u8(0b10000000 | (num & 0b01111111));
                this.u8(0b10000000 | ((num >>> 7) & 0b01111111));
                this.u8(0b10000000 | ((num >>> 14) & 0b01111111));
                this.u8(0b10000000 | ((num >>> 21) & 0b01111111));
                this.u8((hi32 << 4) | (num >>> 28));
            }
            else if (num <= 4398046511103) {
                this.u8(0b10000000 | (num & 0b01111111));
                this.u8(0b10000000 | ((num >>> 7) & 0b01111111));
                this.u8(0b10000000 | ((num >>> 14) & 0b01111111));
                this.u8(0b10000000 | ((num >>> 21) & 0b01111111));
                this.u8(0b10000000 | ((hi32 & 0b111) << 4) | (num >>> 28));
                this.u8(hi32 >>> 3);
            }
            else if (num <= 562949953421311) {
                this.u8(0b10000000 | (num & 0b01111111));
                this.u8(0b10000000 | ((num >>> 7) & 0b01111111));
                this.u8(0b10000000 | ((num >>> 14) & 0b01111111));
                this.u8(0b10000000 | ((num >>> 21) & 0b01111111));
                this.u8(0b10000000 | ((hi32 & 0b111) << 4) | (num >>> 28));
                this.u8(0b10000000 | ((hi32 & 1016) >>> 3));
                this.u8(hi32 >>> 10);
            }
            else {
                this.u8(0b10000000 | (num & 0b01111111));
                this.u8(0b10000000 | ((num >>> 7) & 0b01111111));
                this.u8(0b10000000 | ((num >>> 14) & 0b01111111));
                this.u8(0b10000000 | ((num >>> 21) & 0b01111111));
                this.u8(0b10000000 | ((hi32 & 0b111) << 4) | (num >>> 28));
                this.u8(0b10000000 | ((hi32 & 1016) >>> 3));
                this.u8(0b10000000 | ((hi32 & 130048) >>> 10));
                this.u8(hi32 >>> 17);
            }
        }
    }
    vuint39(num) {
        if (num <= 0b1111111) {
            this.u8(num);
        }
        else if (num <= 16383) {
            this.u8(0b10000000 | (num & 0b01111111));
            this.u8(num >>> 7);
        }
        else if (num <= 2097151) {
            this.u8(0b10000000 | (num & 0b01111111));
            this.u8(0b10000000 | ((num >>> 7) & 0b01111111));
            this.u8(num >>> 14);
        }
        else if (num <= 268435455) {
            this.u8(0b10000000 | (num & 0b01111111));
            this.u8(0b10000000 | ((num >>> 7) & 0b01111111));
            this.u8(0b10000000 | ((num >>> 14) & 0b01111111));
            this.u8(num >>> 21);
        }
        else {
            let lo32 = num | 0;
            if (lo32 < 0)
                lo32 += 4294967296;
            const hi32 = (num - lo32) / 4294967296;
            if (num <= 34359738367) {
                this.u8(0b10000000 | (num & 0b01111111));
                this.u8(0b10000000 | ((num >>> 7) & 0b01111111));
                this.u8(0b10000000 | ((num >>> 14) & 0b01111111));
                this.u8(0b10000000 | ((num >>> 21) & 0b01111111));
                this.u8((hi32 << 4) | (num >>> 28));
            }
            else if (num <= 4398046511103) {
                this.u8(0b10000000 | (num & 0b01111111));
                this.u8(0b10000000 | ((num >>> 7) & 0b01111111));
                this.u8(0b10000000 | ((num >>> 14) & 0b01111111));
                this.u8(0b10000000 | ((num >>> 21) & 0b01111111));
                this.u8(0b10000000 | ((hi32 & 0b111) << 4) | (num >>> 28));
                this.u8((hi32 >>> 3) & 0b1111);
            }
        }
    }
    b1vuint56(flag, num) {
        if (num <= 0b111111) {
            this.u8((flag ? 0b10000000 : 0b00000000) | num);
        }
        else {
            const firstByteMask = flag ? 0b11000000 : 0b01000000;
            if (num <= 8191) {
                this.u8(firstByteMask | (num & 0b00111111));
                this.u8(num >>> 6);
            }
            else if (num <= 1048575) {
                this.u8(firstByteMask | (num & 0b00111111));
                this.u8(0b10000000 | ((num >>> 6) & 0b01111111));
                this.u8(num >>> 13);
            }
            else if (num <= 134217727) {
                this.u8(firstByteMask | (num & 0b00111111));
                this.u8(0b10000000 | ((num >>> 6) & 0b01111111));
                this.u8(0b10000000 | ((num >>> 13) & 0b01111111));
                this.u8(num >>> 20);
            }
            else {
                let lo32 = num | 0;
                if (lo32 < 0)
                    lo32 += 4294967296;
                const hi32 = (num - lo32) / 4294967296;
                if (num <= 17179869183) {
                    this.u8(firstByteMask | (num & 0b00111111));
                    this.u8(0b10000000 | ((num >>> 6) & 0b01111111));
                    this.u8(0b10000000 | ((num >>> 13) & 0b01111111));
                    this.u8(0b10000000 | ((num >>> 20) & 0b01111111));
                    this.u8((hi32 << 5) | (num >>> 27));
                }
                else if (num <= 2199023255551) {
                    this.u8(firstByteMask | (num & 0b00111111));
                    this.u8(0b10000000 | ((num >>> 6) & 0b01111111));
                    this.u8(0b10000000 | ((num >>> 13) & 0b01111111));
                    this.u8(0b10000000 | ((num >>> 20) & 0b01111111));
                    this.u8(0b10000000 | ((hi32 & 0b11) << 5) | (num >>> 27));
                    this.u8(hi32 >>> 2);
                }
                else if (num <= 281474976710655) {
                    this.u8(firstByteMask | (num & 0b00111111));
                    this.u8(0b10000000 | ((num >>> 6) & 0b01111111));
                    this.u8(0b10000000 | ((num >>> 13) & 0b01111111));
                    this.u8(0b10000000 | ((num >>> 20) & 0b01111111));
                    this.u8(0b10000000 | ((hi32 & 0b11) << 5) | (num >>> 27));
                    this.u8(0b10000000 | ((hi32 & 508) >>> 2));
                    this.u8(hi32 >>> 9);
                }
                else {
                    this.u8(firstByteMask | (num & 0b00111111));
                    this.u8(0b10000000 | ((num >>> 6) & 0b01111111));
                    this.u8(0b10000000 | ((num >>> 13) & 0b01111111));
                    this.u8(0b10000000 | ((num >>> 20) & 0b01111111));
                    this.u8(0b10000000 | ((hi32 & 0b11) << 5) | (num >>> 27));
                    this.u8(0b10000000 | ((hi32 & 508) >>> 2));
                    this.u8(0b10000000 | ((hi32 & 65024) >>> 9));
                    this.u8(hi32 >>> 16);
                }
            }
        }
    }
    b1vuint28(flag, num) {
        if (num <= 0b111111) {
            this.u8((flag ? 0b10000000 : 0b00000000) | num);
        }
        else {
            const firstByteMask = flag ? 0b11000000 : 0b01000000;
            if (num <= 8191) {
                this.u8(firstByteMask | (num & 0b00111111));
                this.u8(num >>> 6);
            }
            else if (num <= 1048575) {
                this.u8(firstByteMask | (num & 0b00111111));
                this.u8(0b10000000 | ((num >>> 6) & 0b01111111));
                this.u8(num >>> 13);
            }
            else {
                this.u8(firstByteMask | (num & 0b00111111));
                this.u8(0b10000000 | ((num >>> 6) & 0b01111111));
                this.u8(0b10000000 | ((num >>> 13) & 0b01111111));
                this.u8(num >>> 20);
            }
        }
    }
    uint53vuint39(x, z) {
        let x1 = x | 0;
        if (x1 < 0)
            x1 += 4294967296;
        const x2 = (x - x1) / 4294967296;
        this.u32(x1);
        this.u16(x2 & 0xffff);
        const fiveXBits = (x2 >>> 16) << 3;
        const twoZBits = (z >>> 8) & 0b11;
        const zFitsIn10Bits = z <= 1023;
        this.u8(fiveXBits | (zFitsIn10Bits ? 0b000 : 0b100) | twoZBits);
        this.u8(z & 0xff);
        if (zFitsIn10Bits)
            return;
        if (z <= 131071) {
            this.u8(z >>> 10);
        }
        else if (z <= 16777215) {
            this.u8(128 | ((z >>> 10) & 127));
            this.u8(z >>> 17);
        }
        else if (z <= 2147483647) {
            this.u8(128 | ((z >>> 10) & 127));
            this.u8(128 | ((z >>> 17) & 127));
            this.u8(z >>> 24);
        }
        else {
            let z1 = z | 0;
            if (z1 < 0)
                z1 += 4294967296;
            const z2 = (z - z1) / 4294967296;
            this.u8(128 | ((z1 >>> 10) & 127));
            this.u8(128 | ((z1 >>> 17) & 127));
            this.u8(128 | ((z1 >>> 24) & 127));
            this.u8((z1 >>> 31) | (z2 << 1));
        }
    }
}
exports.AbstractEncoder = AbstractEncoder;
