"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ServerDecoder = void 0;
const clock_1 = require("../../../json-crdt-patch/clock");
const model_1 = require("../../model");
const DocRootType_1 = require("../../types/lww-doc-root/DocRootType");
const constants_1 = require("../../../json-crdt-patch/constants");
const constants_2 = require("../../constants");
const ConstantType_1 = require("../../types/const/ConstantType");
const ObjectChunk_1 = require("../../types/lww-object/ObjectChunk");
const ObjectType_1 = require("../../types/lww-object/ObjectType");
const ValueType_1 = require("../../types/lww-value/ValueType");
const ArrayChunk_1 = require("../../types/rga-array/ArrayChunk");
const ArrayType_1 = require("../../types/rga-array/ArrayType");
const StringChunk_1 = require("../../types/rga-string/StringChunk");
const StringType_1 = require("../../types/rga-string/StringType");
class ServerDecoder {
    decode(data) {
        this.time = data[0];
        const doc = model_1.Model.withServerClock(this.time);
        const rootId = this.ts(data, 1);
        const rootNode = data[2] ? this.decodeNode(doc, data[2]) : null;
        doc.root = new DocRootType_1.DocRootType(doc, rootId, rootNode);
        return doc;
    }
    ts(arr, index) {
        const timeDiff = arr[index];
        return new clock_1.ServerTimestamp(this.time - timeDiff);
    }
    decodeNode(doc, data) {
        switch (data) {
            case 1:
                return constants_2.NULL;
            case 2:
                return constants_2.FALSE;
            case 3:
                return constants_2.TRUE;
            case 4:
                return constants_2.UNDEFINED;
        }
        if (data instanceof Array) {
            switch (data[0]) {
                case 0:
                    return this.decodeObj(doc, data);
                case 1:
                    return this.decodeArr(doc, data);
                case 2:
                    return this.decodeStr(doc, data);
                case 3:
                    return this.decodeVal(doc, data);
                case 4:
                    return this.decodeConst(doc, data);
            }
        }
        throw new Error('UNKNOWN_NODE');
    }
    decodeObj(doc, data) {
        const id = this.ts(data, 1);
        const obj = new ObjectType_1.ObjectType(doc, id);
        const length = data.length;
        for (let i = 2; i < length; i += 3) {
            const key = data[i];
            const keyId = this.ts(data, i + 1);
            const node = this.decodeNode(doc, data[i + 2]);
            const chunk = new ObjectChunk_1.ObjectChunk(keyId, node);
            obj.putChunk(key, chunk);
        }
        doc.nodes.index(obj);
        return obj;
    }
    decodeArr(doc, data) {
        const id = this.ts(data, 1);
        const obj = new ArrayType_1.ArrayType(doc, id);
        const length = data.length;
        for (let i = 2; i < length; i += 2) {
            const chunkId = this.ts(data, i);
            const content = data[i + 1];
            let chunk;
            if (typeof content === 'number') {
                chunk = new ArrayChunk_1.ArrayChunk(chunkId, undefined);
                chunk.deleted = content;
            }
            else {
                const nodes = content.map((c) => this.decodeNode(doc, c));
                chunk = new ArrayChunk_1.ArrayChunk(chunkId, nodes);
            }
            obj.append(chunk);
        }
        doc.nodes.index(obj);
        return obj;
    }
    decodeStr(doc, data) {
        const id = this.ts(data, 1);
        const obj = new StringType_1.StringType(doc, id);
        const length = data.length;
        for (let i = 2; i < length; i += 2) {
            const chunkId = this.ts(data, i);
            const content = data[i + 1];
            let chunk;
            if (typeof content === 'number') {
                chunk = new StringChunk_1.StringChunk(chunkId, undefined);
                chunk.deleted = content;
            }
            else {
                chunk = new StringChunk_1.StringChunk(chunkId, content);
            }
            obj.append(chunk);
        }
        doc.nodes.index(obj);
        return obj;
    }
    decodeVal(doc, data) {
        const id = this.ts(data, 1);
        const writeId = this.ts(data, 2);
        const value = data[3];
        const obj = new ValueType_1.ValueType(id, writeId, value);
        doc.nodes.index(obj);
        return obj;
    }
    decodeConst(doc, data) {
        const value = data[1];
        const obj = new ConstantType_1.ConstantType(constants_1.ORIGIN, value);
        return obj;
    }
}
exports.ServerDecoder = ServerDecoder;
