"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.AbstractDecoder = void 0;
const constants_1 = require("../../../json-crdt-patch/constants");
const constants_2 = require("../../constants");
const ConstantType_1 = require("../../types/const/ConstantType");
const DocRootType_1 = require("../../types/lww-doc-root/DocRootType");
const ObjectChunk_1 = require("../../types/lww-object/ObjectChunk");
const ObjectType_1 = require("../../types/lww-object/ObjectType");
const ValueType_1 = require("../../types/lww-value/ValueType");
const ArrayChunk_1 = require("../../types/rga-array/ArrayChunk");
const ArrayType_1 = require("../../types/rga-array/ArrayType");
const StringChunk_1 = require("../../types/rga-string/StringChunk");
const StringType_1 = require("../../types/rga-string/StringType");
class AbstractDecoder {
    decodeRoot(doc, { id, node }) {
        const ts = this.decodeTimestamp(id);
        const jsonNode = node ? this.decodeNode(doc, node) : null;
        const root = new DocRootType_1.DocRootType(doc, ts, jsonNode);
        doc.root = root;
    }
    decodeNode(doc, node) {
        switch (node.type) {
            case 'obj':
                return this.decodeObj(doc, node);
            case 'arr':
                return this.decodeArr(doc, node);
            case 'str':
                return this.decodeStr(doc, node);
            case 'val':
                return this.decodeVal(doc, node);
            case 'const':
                return this.decodeConst(doc, node);
        }
        throw new Error('UNKNOWN_NODE');
    }
    decodeObj(doc, node) {
        const id = this.decodeTimestamp(node.id);
        const obj = new ObjectType_1.ObjectType(doc, id);
        const keys = Object.keys(node.chunks);
        for (const key of keys) {
            const val = node.chunks[key];
            obj.putChunk(key, new ObjectChunk_1.ObjectChunk(this.decodeTimestamp(val.id), this.decodeNode(doc, val.node)));
        }
        doc.nodes.index(obj);
        return obj;
    }
    decodeArr(doc, node) {
        const obj = new ArrayType_1.ArrayType(doc, this.decodeTimestamp(node.id));
        for (const c of node.chunks)
            obj.append(this.decodeArrChunk(doc, c));
        doc.nodes.index(obj);
        return obj;
    }
    decodeArrChunk(doc, c) {
        const id = this.decodeTimestamp(c.id);
        if (typeof c.span === 'number') {
            const chunk = new ArrayChunk_1.ArrayChunk(id, undefined);
            chunk.deleted = c.span;
            return chunk;
        }
        else
            return new ArrayChunk_1.ArrayChunk(id, c.nodes.map((n) => this.decodeNode(doc, n)));
    }
    decodeStr(doc, node) {
        const obj = new StringType_1.StringType(doc, this.decodeTimestamp(node.id));
        for (const c of node.chunks)
            obj.append(this.decodeStrChunk(doc, c));
        doc.nodes.index(obj);
        return obj;
    }
    decodeStrChunk(doc, c) {
        const id = this.decodeTimestamp(c.id);
        if (typeof c.span === 'number') {
            const chunk = new StringChunk_1.StringChunk(id, undefined);
            chunk.deleted = c.span;
            return chunk;
        }
        else
            return new StringChunk_1.StringChunk(id, c.value);
    }
    decodeVal(doc, node) {
        const obj = new ValueType_1.ValueType(this.decodeTimestamp(node.id), this.decodeTimestamp(node.writeId), node.value);
        doc.nodes.index(obj);
        return obj;
    }
    decodeConst(doc, node) {
        switch (node.value) {
            case null:
                return constants_2.NULL;
            case true:
                return constants_2.TRUE;
            case false:
                return constants_2.FALSE;
            case undefined:
                return constants_2.UNDEFINED;
        }
        return new ConstantType_1.ConstantType(constants_1.ORIGIN, node.value);
    }
}
exports.AbstractDecoder = AbstractDecoder;
