"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.AbstractEncoder = void 0;
const ConstantType_1 = require("../../types/const/ConstantType");
const ObjectType_1 = require("../../types/lww-object/ObjectType");
const ValueType_1 = require("../../types/lww-value/ValueType");
const ArrayType_1 = require("../../types/rga-array/ArrayType");
const StringType_1 = require("../../types/rga-string/StringType");
class AbstractEncoder {
    encodeRoot(root) {
        return {
            type: 'root',
            id: this.encodeTimestamp(root.id),
            node: root.node ? this.encodeNode(root.node) : null,
        };
    }
    encodeNode(node) {
        if (node instanceof ObjectType_1.ObjectType)
            return this.encodeObj(node);
        else if (node instanceof ArrayType_1.ArrayType)
            return this.encodeArr(node);
        else if (node instanceof StringType_1.StringType)
            return this.encodeStr(node);
        else if (node instanceof ValueType_1.ValueType)
            return this.encodeVal(node);
        else if (node instanceof ConstantType_1.ConstantType)
            return this.encodeConst(node);
        throw new Error('UNKNOWN_NODE');
    }
    encodeObj(obj) {
        const chunks = {};
        for (const [key, entry] of obj.latest.entries())
            chunks[key] = {
                id: this.encodeTimestamp(entry.id),
                node: this.encodeNode(entry.node),
            };
        return {
            type: 'obj',
            id: this.encodeTimestamp(obj.id),
            chunks,
        };
    }
    encodeArr(obj) {
        const chunks = [];
        for (const chunk of obj.chunks())
            chunks.push(this.encodeArrChunk(chunk));
        return {
            type: 'arr',
            id: this.encodeTimestamp(obj.id),
            chunks,
        };
    }
    encodeArrChunk(chunk) {
        if (chunk.deleted) {
            const tombstone = {
                id: this.encodeTimestamp(chunk.id),
                span: chunk.deleted,
            };
            return tombstone;
        }
        const res = {
            id: this.encodeTimestamp(chunk.id),
            nodes: chunk.nodes.map((n) => this.encodeNode(n)),
        };
        return res;
    }
    encodeStr(obj) {
        const chunks = [];
        for (const chunk of obj.chunks())
            chunks.push(this.encodeStrChunk(chunk));
        return {
            type: 'str',
            id: this.encodeTimestamp(obj.id),
            chunks,
        };
    }
    encodeStrChunk(chunk) {
        if (chunk.deleted) {
            const tombstone = {
                id: this.encodeTimestamp(chunk.id),
                span: chunk.deleted,
            };
            return tombstone;
        }
        const res = {
            id: this.encodeTimestamp(chunk.id),
            value: chunk.str,
        };
        return res;
    }
    encodeVal(obj) {
        return {
            type: 'val',
            id: this.encodeTimestamp(obj.id),
            writeId: this.encodeTimestamp(obj.writeId),
            value: obj.value,
        };
    }
    encodeConst(obj) {
        return {
            type: 'const',
            value: obj.value,
        };
    }
}
exports.AbstractEncoder = AbstractEncoder;
