"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JsonPatchDraft = void 0;
const Draft_1 = require("../../json-crdt-patch/Draft");
const op_1 = require("../../json-patch/op");
const ObjectType_1 = require("../types/lww-object/ObjectType");
const ArrayType_1 = require("../types/rga-array/ArrayType");
const constants_1 = require("../../json-crdt-patch/constants");
const json_pointer_1 = require("../../json-pointer");
const isEqual = require('fast-deep-equal');
class JsonPatchDraft extends Draft_1.Draft {
    constructor(model) {
        super();
        this.model = model;
    }
    applyOps(ops) {
        for (const op of ops)
            this.applyOp(op);
    }
    applyOp(op) {
        if (op instanceof op_1.OpAdd)
            this.applyOpAdd(op);
        else if (op instanceof op_1.OpRemove)
            this.applyOpRemove(op);
        else if (op instanceof op_1.OpReplace)
            this.applyOpReplace(op);
        else if (op instanceof op_1.OpMove)
            this.applyOpMove(op);
        else if (op instanceof op_1.OpCopy)
            this.applyOpCopy(op);
        else if (op instanceof op_1.OpTest)
            this.applyOpTest(op);
    }
    applyOpAdd(op) {
        const { builder } = this;
        const steps = op.path;
        if (!steps.length)
            this.setRoot(op.value);
        else {
            const objSteps = steps.slice(0, steps.length - 1);
            const node = this.model.api.find(objSteps);
            const key = steps[steps.length - 1];
            if (node instanceof ObjectType_1.ObjectType) {
                builder.setKeys(node.id, [[String(key), builder.json(op.value)]]);
            }
            else if (node instanceof ArrayType_1.ArrayType) {
                const value = builder.json(op.value);
                if (key === '-') {
                    const length = node.size();
                    const after = node.findId(length);
                    builder.insArr(node.id, after, [value]);
                }
                else {
                    const index = ~~key;
                    if ('' + index !== key)
                        throw new Error('INVALID_INDEX');
                    if (!index)
                        builder.insArr(node.id, node.id, [value]);
                    else {
                        const after = node.findId(index - 1);
                        builder.insArr(node.id, after, [value]);
                    }
                }
            }
            else
                throw new Error('NOT_FOUND');
        }
    }
    applyOpRemove(op) {
        const { builder } = this;
        const steps = op.path;
        if (!steps.length)
            this.setRoot(null);
        else {
            const objSteps = steps.slice(0, steps.length - 1);
            const node = this.model.api.find(objSteps);
            const key = steps[steps.length - 1];
            if (node instanceof ObjectType_1.ObjectType) {
                const stringKey = String(key);
                if (node.get(stringKey) === undefined)
                    throw new Error('NOT_FOUND');
                builder.setKeys(node.id, [[stringKey, constants_1.UNDEFINED_ID]]);
            }
            else if (node instanceof ArrayType_1.ArrayType) {
                const key = steps[steps.length - 1];
                const index = ~~key;
                if ('' + index !== key)
                    throw new Error('INVALID_INDEX');
                const id = node.findId(index);
                builder.del(node.id, id, 1);
            }
            else
                throw new Error('NOT_FOUND');
        }
    }
    applyOpReplace(op) {
        const { path, value } = op;
        this.applyOpRemove(new op_1.OpRemove(path, undefined));
        this.applyOpAdd(new op_1.OpAdd(path, value));
    }
    applyOpMove(op) {
        const { path, from } = op;
        if (json_pointer_1.isChild(from, path))
            throw new Error('INVALID_CHILD');
        const json = this.json(from);
        this.applyOpRemove(new op_1.OpRemove(from, undefined));
        this.applyOpAdd(new op_1.OpAdd(path, json));
    }
    applyOpCopy(op) {
        const { path, from } = op;
        const json = this.json(from);
        this.applyOpAdd(new op_1.OpAdd(path, json));
    }
    applyOpTest(op) {
        const { path, value } = op;
        const json = this.json(path);
        if (!isEqual(json, value))
            throw new Error('TEST');
    }
    get(steps) {
        var _a, _b;
        if (!steps.length)
            return this.model.toJson();
        else {
            const objSteps = steps.slice(0, steps.length - 1);
            const node = this.model.api.find(objSteps);
            const key = steps[steps.length - 1];
            if (node instanceof ObjectType_1.ObjectType) {
                return (_a = node.getNode(String(key))) === null || _a === void 0 ? void 0 : _a.toJson();
            }
            else if (node instanceof ArrayType_1.ArrayType) {
                const index = ~~key;
                if ('' + index !== key)
                    throw new Error('INVALID_INDEX');
                const nodeId = node.findValue(index);
                return (_b = this.model.node(nodeId)) === null || _b === void 0 ? void 0 : _b.toJson();
            }
        }
        return undefined;
    }
    json(steps) {
        const json = this.get(steps);
        if (json === undefined)
            throw new Error('NOT_FOUND');
        return json;
    }
    setRoot(json) {
        const { builder } = this;
        builder.root(builder.json(json));
    }
}
exports.JsonPatchDraft = JsonPatchDraft;
