"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Model = void 0;
const constants_1 = require("../constants");
const nodes_1 = require("./nodes");
const util_1 = require("./util");
const SetRootOperation_1 = require("../../json-crdt-patch/operations/SetRootOperation");
const clock_1 = require("../../json-crdt-patch/clock");
const DocRootType_1 = require("../types/lww-doc-root/DocRootType");
const ObjectType_1 = require("../types/lww-object/ObjectType");
const ArrayType_1 = require("../types/rga-array/ArrayType");
const StringType_1 = require("../types/rga-string/StringType");
const MakeObjectOperation_1 = require("../../json-crdt-patch/operations/MakeObjectOperation");
const SetObjectKeysOperation_1 = require("../../json-crdt-patch/operations/SetObjectKeysOperation");
const MakeArrayOperation_1 = require("../../json-crdt-patch/operations/MakeArrayOperation");
const InsertArrayElementsOperation_1 = require("../../json-crdt-patch/operations/InsertArrayElementsOperation");
const constants_2 = require("../../json-crdt-patch/constants");
const DeleteOperation_1 = require("../../json-crdt-patch/operations/DeleteOperation");
const MakeStringOperation_1 = require("../../json-crdt-patch/operations/MakeStringOperation");
const InsertStringSubstringOperation_1 = require("../../json-crdt-patch/operations/InsertStringSubstringOperation");
const ModelApi_1 = require("./api/ModelApi");
const MakeValueOperation_1 = require("../../json-crdt-patch/operations/MakeValueOperation");
const ValueType_1 = require("../types/lww-value/ValueType");
const SetValueOperation_1 = require("../../json-crdt-patch/operations/SetValueOperation");
class Model {
    constructor(clock, nodes) {
        this.root = new DocRootType_1.DocRootType(this, constants_2.ORIGIN, null);
        this.clock = clock;
        this.nodes = nodes;
        this.api = new ModelApi_1.ModelApi(this);
    }
    static withLogicalClock(clock) {
        clock = clock || new clock_1.LogicalVectorClock(util_1.randomSessionId(), 0);
        const nodes = new nodes_1.LogicalNodeIndex();
        return new Model(clock, nodes);
    }
    static withServerClock(time) {
        const clock = new clock_1.ServerVectorClock(time || 0);
        const nodes = new nodes_1.ServerNodeIndex();
        return new Model(clock, nodes);
    }
    node(id) {
        if (id.getSessionId() === 0) {
            switch (id.time) {
                case 1: return constants_1.NULL;
                case 2: return constants_1.TRUE;
                case 3: return constants_1.FALSE;
                case 4: return constants_1.UNDEFINED;
            }
        }
        return this.nodes.get(id);
    }
    applyPatch(patch) {
        const ops = patch.ops;
        const { length } = ops;
        for (let i = 0; i < length; i++)
            this.applyOperation(ops[i]);
    }
    applyOperation(op) {
        this.clock.observe(op.id, op.span());
        if (op instanceof MakeObjectOperation_1.MakeObjectOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ObjectType_1.ObjectType(this, op.id));
        }
        else if (op instanceof MakeArrayOperation_1.MakeArrayOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ArrayType_1.ArrayType(this, op.id));
        }
        else if (op instanceof MakeStringOperation_1.MakeStringOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new StringType_1.StringType(this, op.id));
        }
        else if (op instanceof MakeValueOperation_1.MakeValueOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ValueType_1.ValueType(op.id, op.id, op.value));
        }
        else if (op instanceof SetRootOperation_1.SetRootOperation) {
            const oldValue = this.root.insert(op);
            if (oldValue)
                this.deleteNodeTree(oldValue);
        }
        else if (op instanceof SetObjectKeysOperation_1.SetObjectKeysOperation) {
            const obj = this.nodes.get(op.object);
            if (obj instanceof ObjectType_1.ObjectType)
                obj.insert(op);
        }
        else if (op instanceof SetValueOperation_1.SetValueOperation) {
            const obj = this.nodes.get(op.obj);
            if (obj instanceof ValueType_1.ValueType)
                obj.insert(op);
        }
        else if (op instanceof InsertArrayElementsOperation_1.InsertArrayElementsOperation) {
            const arr = this.nodes.get(op.arr);
            if (arr instanceof ArrayType_1.ArrayType)
                arr.insert(op);
        }
        else if (op instanceof InsertStringSubstringOperation_1.InsertStringSubstringOperation) {
            const arr = this.nodes.get(op.obj);
            if (arr instanceof StringType_1.StringType)
                arr.onInsert(op);
        }
        else if (op instanceof DeleteOperation_1.DeleteOperation) {
            const node = this.nodes.get(op.obj);
            if (node instanceof ArrayType_1.ArrayType)
                node.delete(op);
            else if (node instanceof StringType_1.StringType)
                node.onDelete(op);
        }
    }
    deleteNodeTree(value) {
        const isSystemNode = value.getSessionId() < 1;
        if (isSystemNode)
            return;
        const node = this.nodes.get(value);
        if (!node)
            return;
        for (const child of node.children())
            this.deleteNodeTree(child);
        this.nodes.delete(value);
    }
    clone() {
        const model = this.clock instanceof clock_1.LogicalVectorClock
            ? Model.withLogicalClock(this.clock.clone())
            : Model.withServerClock(this.clock.time);
        model.root = this.root.clone(model);
        return model;
    }
    fork(sessionId = util_1.randomSessionId()) {
        const model = this.clock instanceof clock_1.LogicalVectorClock
            ? Model.withLogicalClock(this.clock.fork(sessionId))
            : Model.withServerClock(this.clock.time);
        model.root = this.root.clone(model);
        return model;
    }
    toJson() {
        return this.root.toJson();
    }
    toString(tab) {
        return this.root.toString(tab);
    }
}
exports.Model = Model;
