"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ArrayType = void 0;
const ArrayChunk_1 = require("./ArrayChunk");
const ArrayOriginChunk_1 = require("./ArrayOriginChunk");
class ArrayType {
    constructor(doc, id) {
        this.doc = doc;
        this.id = id;
        this.start = this.end = new ArrayOriginChunk_1.ArrayOriginChunk(id);
    }
    insert(op) {
        let curr = this.end;
        while (curr) {
            if (curr.id.overlap(curr.span(), op.id, op.span()))
                return;
            if (curr.id.inSpan(curr.span(), op.after, 1))
                break;
            curr = curr.left;
        }
        if (!curr)
            return;
        const nodes = [];
        for (const el of op.elements) {
            const node = this.doc.node(el);
            if (node)
                nodes.push(node);
        }
        if (!nodes.length)
            return;
        const isOriginChunk = curr instanceof ArrayOriginChunk_1.ArrayOriginChunk;
        if (!curr.deleted && !isOriginChunk) {
            const doesAfterMatch = curr.id.getSessionId() === op.after.getSessionId() && curr.id.time + curr.span() - 1 === op.after.time;
            const isIdSameSession = curr.id.getSessionId() === op.id.getSessionId();
            const isIdIncreasingWithoutAGap = curr.id.time + curr.span() === op.id.time;
            if (doesAfterMatch && isIdSameSession && isIdIncreasingWithoutAGap) {
                curr.merge(nodes);
                return;
            }
        }
        const chunk = new ArrayChunk_1.ArrayChunk(op.id, nodes);
        const targetsLastElementInChunk = op.after.time === curr.id.time + curr.span() - 1;
        if (targetsLastElementInChunk) {
            while (curr.right && curr.right.id.compare(op.id) > 0)
                curr = curr.right;
            this.insertChunk(chunk, curr);
            return;
        }
        this.splitChunk(curr, op.after.time);
        this.insertChunk(chunk, curr);
    }
    insertChunk(chunk, after) {
        chunk.left = after;
        chunk.right = after.right;
        if (after.right)
            after.right.left = chunk;
        else
            this.end = chunk;
        after.right = chunk;
    }
    splitChunk(chunk, time) {
        const newChunk = chunk.split(time);
        this.insertChunk(newChunk, chunk);
        return newChunk;
    }
    delete(op) {
        const { after, length } = op;
        let chunk = this.end;
        const chunks = [];
        while (chunk) {
            if (chunk.id.overlap(chunk.span(), after, length))
                chunks.push(chunk);
            if (chunk.id.inSpan(chunk.span(), after, 1))
                break;
            chunk = chunk.left;
        }
        for (const c of chunks)
            this.deleteInChunk(c, after.time, after.time + length - 1);
    }
    deleteInChunk(chunk, t1, t2) {
        const c1 = chunk.id.time;
        const c2 = c1 + chunk.span() - 1;
        if (t1 <= c1) {
            if (t2 >= c2)
                chunk.delete();
            else {
                this.splitChunk(chunk, t2);
                chunk.delete();
            }
        }
        else {
            if (t2 >= c2)
                this.splitChunk(chunk, t1 - 1).delete();
            else {
                this.splitChunk(chunk, t2);
                this.splitChunk(chunk, t1 - 1).delete();
            }
        }
    }
    findId(index) {
        let chunk = this.start;
        let cnt = 0;
        const next = index + 1;
        while (chunk) {
            if (chunk.nodes) {
                cnt += chunk.nodes.length;
                if (cnt >= next)
                    return chunk.id.tick(chunk.nodes.length - (cnt - index));
            }
            chunk = chunk.right;
        }
        throw new Error('OUT_OF_BOUNDS');
    }
    findValue(index) {
        let chunk = this.start;
        let cnt = 0;
        const next = index + 1;
        while (chunk) {
            if (chunk.nodes) {
                cnt += chunk.nodes.length;
                if (cnt >= next)
                    return chunk.nodes[chunk.nodes.length - (cnt - index)].id;
            }
            chunk = chunk.right;
        }
        throw new Error('OUT_OF_BOUNDS');
    }
    findIdSpans(index, length) {
        let chunk = this.start;
        let cnt = 0;
        const next = index + 1;
        while (chunk) {
            if (chunk.nodes) {
                cnt += chunk.nodes.length;
                if (cnt >= next) {
                    const remaining = cnt - index;
                    if (remaining >= length)
                        return [chunk.id.interval(chunk.span() - remaining, length)];
                    length -= remaining;
                    const result = [chunk.id.interval(chunk.span() - remaining, remaining)];
                    while (chunk.right) {
                        chunk = chunk.right;
                        if (chunk.deleted)
                            continue;
                        const span = chunk.span();
                        if (span >= length) {
                            result.push(chunk.id.interval(0, length));
                            break;
                        }
                        result.push(chunk.id.interval(0, span));
                        length -= span;
                    }
                    return result;
                }
            }
            chunk = chunk.right;
        }
        throw new Error('OUT_OF_BOUNDS');
    }
    append(chunk) {
        const last = this.end;
        last.right = chunk;
        chunk.left = last;
        this.end = chunk;
    }
    toJson() {
        const arr = [];
        const nodes = this.doc.nodes;
        let curr = this.start;
        while (curr) {
            if (curr.nodes)
                arr.push(...curr.nodes.map((node) => node.toJson()));
            curr = curr.right;
        }
        return arr;
    }
    clone(model) {
        const copy = new ArrayType(model, this.id);
        let i = this.start;
        let j = copy.start;
        while (i.right) {
            const cloned = i.right.clone(model);
            j.right = cloned;
            cloned.left = j;
            j = cloned;
            i = i.right;
        }
        copy.end = j;
        model.nodes.index(copy);
        return copy;
    }
    *children() {
        let chunk = this.start;
        while ((chunk = chunk.right))
            if (chunk.nodes)
                for (const node of chunk.nodes)
                    yield node.id;
    }
    *chunks() {
        let curr = this.start;
        while ((curr = curr.right))
            yield curr;
    }
    size() {
        let curr = this.start;
        let size = 0;
        while ((curr = curr.right))
            size++;
        return size;
    }
    length() {
        let curr = this.start;
        let size = 0;
        while ((curr = curr.right))
            if (curr.nodes)
                size += curr.nodes.length;
        return size;
    }
    toString(tab = '') {
        let str = `${tab}ArrayType(${this.id.toDisplayString()})`;
        let curr = this.start;
        while (curr) {
            str += `\n${curr.toString(tab + '  ')}`;
            curr = curr.right;
        }
        return str;
    }
}
exports.ArrayType = ArrayType;
