"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RpcApiCaller = void 0;
const tslib_1 = require("tslib");
const rxjs_1 = require("rxjs");
const operators_1 = require("rxjs/operators");
const constants_1 = require("./constants");
const BufferSubject_1 = require("../../../util/BufferSubject");
const error_1 = require("./error");
class RpcApiCaller {
    constructor({ api, error, preCallBufferSize = 10, maxActiveCalls = 50, }) {
        this._calls = 0;
        this.api = api;
        this.error = error || new error_1.ErrorLikeErrorFormatter();
        this.preCallBufferSize = preCallBufferSize;
        this.maxActiveCalls = maxActiveCalls;
    }
    get activeCalls() {
        return this._calls;
    }
    exists(name) {
        return this.api.hasOwnProperty(name);
    }
    get(name) {
        if (!this.exists(name))
            throw new error_1.RpcError(constants_1.RpcServerError.NoMethodSpecified);
        return this.api[name];
    }
    call(name, request, ctx) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            try {
                if (this._calls >= this.maxActiveCalls)
                    throw new error_1.RpcError(constants_1.RpcServerError.TooManyActiveCalls);
                const method = this.get(name);
                if (method.validate)
                    method.validate(request);
                try {
                    this._calls++;
                    if (method.onPreCall)
                        yield method.onPreCall(ctx, request);
                    return !method.isStreaming
                        ? yield method.call(ctx, request)
                        : yield rxjs_1.firstValueFrom(method.call$(ctx, rxjs_1.of(request)));
                }
                finally {
                    this._calls--;
                }
            }
            catch (error) {
                throw this.error.format(error);
            }
        });
    }
    createCall(name, ctx) {
        const req$ = new rxjs_1.Subject();
        const reqUnsubscribe$ = new rxjs_1.Subject();
        try {
            if (this._calls >= this.maxActiveCalls)
                throw new error_1.RpcError(constants_1.RpcServerError.TooManyActiveCalls);
            const method = this.get(name);
            if (!method.isStreaming) {
                const response$ = rxjs_1.from((() => tslib_1.__awaiter(this, void 0, void 0, function* () {
                    const request = yield rxjs_1.firstValueFrom(req$.pipe(operators_1.first()));
                    const response = yield this.call(name, request, ctx);
                    return response;
                }))());
                const res$ = new rxjs_1.Subject();
                response$.subscribe(res$);
                const $resWithErrorsFormatted = res$.pipe(operators_1.catchError(error => {
                    throw this.error.format(error);
                }));
                return { req$, reqUnsubscribe$, res$: $resWithErrorsFormatted };
            }
            const methodStreaming = method;
            const requestValidated$ = req$
                .pipe(operators_1.map(request => {
                if (methodStreaming.validate)
                    methodStreaming.validate(request);
                return request;
            }));
            const bufferSize = methodStreaming.preCallBufferSize || this.preCallBufferSize;
            const requestBuffered$ = new BufferSubject_1.BufferSubject(bufferSize);
            requestBuffered$.subscribe({ error: () => { } });
            const error$ = new rxjs_1.Subject();
            requestBuffered$.subscribe({
                error: error => { error$.error(error); },
            });
            requestValidated$.subscribe(requestBuffered$);
            const result$ = requestBuffered$
                .pipe(operators_1.take(1), operators_1.switchMap(request => {
                return methodStreaming.onPreCall
                    ? rxjs_1.from(methodStreaming.onPreCall(ctx, request))
                    : rxjs_1.from([0]);
            }), operators_1.switchMap(() => {
                Promise.resolve().then(() => {
                    requestBuffered$.flush();
                });
                return methodStreaming.call$(ctx, requestBuffered$)
                    .pipe(operators_1.finalize(() => {
                    error$.complete();
                }));
            }), operators_1.share(), operators_1.mergeWith(error$));
            const resultWithActiveCallTracking$ = rxjs_1.of(null).pipe(operators_1.tap(() => {
                this._calls++;
            }), operators_1.switchMap(() => result$), operators_1.finalize(() => {
                error$.complete();
                this._calls--;
            }));
            const res$ = new rxjs_1.Observable(observer => {
                var _a;
                const subscription = resultWithActiveCallTracking$.subscribe(observer);
                const timeout = (_a = methodStreaming.timeout) !== null && _a !== void 0 ? _a : 15000;
                const timeoutSubscription = requestBuffered$
                    .pipe(operators_1.catchError(() => rxjs_1.EMPTY), operators_1.mergeWith(result$.pipe(operators_1.catchError(() => rxjs_1.EMPTY))), operators_1.debounce(() => rxjs_1.interval(timeout)))
                    .subscribe(() => {
                    const error = new error_1.RpcError(constants_1.RpcServerError.Timeout);
                    error$.error(error);
                });
                return () => {
                    timeoutSubscription.unsubscribe();
                    subscription.unsubscribe();
                };
            });
            const $resWithErrorsFormatted = res$.pipe(operators_1.catchError(error => {
                throw this.error.format(error);
            }));
            return { req$, reqUnsubscribe$, res$: $resWithErrorsFormatted };
        }
        catch (error) {
            const errorFormatted = this.error.format(error);
            req$.error(errorFormatted);
            const res$ = new rxjs_1.Subject();
            res$.error(errorFormatted);
            return { req$, reqUnsubscribe$, res$ };
        }
    }
    call$(name, request$, ctx) {
        const call = this.createCall(name, ctx);
        request$.subscribe(call.req$);
        return call.res$;
    }
}
exports.RpcApiCaller = RpcApiCaller;
