"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RpcServer = exports.RpcServerError = void 0;
const nominal_1 = require("../messages/nominal");
const subscribeCompleteObserver_1 = require("../util/subscribeCompleteObserver");
const TimedQueue_1 = require("../util/TimedQueue");
const constants_1 = require("./constants");
Object.defineProperty(exports, "RpcServerError", { enumerable: true, get: function () { return constants_1.RpcServerError; } });
const error_1 = require("./error");
class RpcServer {
    constructor({ caller, error, send, onNotification: notify, bufferSize = 10, bufferTime = 1, }) {
        this.activeStreamCalls = new Map();
        this.caller = caller;
        this.error = error || new error_1.ErrorLikeErrorFormatter();
        this.onNotification = notify;
        this.onSend = send;
        if (bufferTime) {
            const buffer = new TimedQueue_1.TimedQueue();
            buffer.itemLimit = bufferSize;
            buffer.timeLimit = bufferTime;
            buffer.onFlush = messages => this.onSend(messages);
            this.send = (message) => {
                buffer.push(message);
            };
        }
        else {
            this.send = message => {
                this.onSend([message]);
            };
        }
    }
    onMessage(message, ctx) {
        try {
            if (message instanceof nominal_1.RequestDataMessage)
                this.onRequestDataMessage(message, ctx);
            else if (message instanceof nominal_1.RequestCompleteMessage)
                this.onRequestCompleteMessage(message, ctx);
            else if (message instanceof nominal_1.RequestErrorMessage)
                this.onRequestErrorMessage(message, ctx);
            else if (message instanceof nominal_1.NotificationMessage)
                this.onNotificationMessage(message, ctx);
            else if (message instanceof nominal_1.ResponseUnsubscribeMessage)
                this.onUnsubscribeMessage(message);
        }
        catch (error) {
            const formattedError = this.error.format(error);
            const message = new nominal_1.NotificationMessage('.err', formattedError);
            this.send(message);
        }
    }
    onMessages(messages, ctx) {
        const length = messages.length;
        for (let i = 0; i < length; i++)
            this.onMessage(messages[i], ctx);
    }
    stop(reason = constants_1.RpcServerError.Stop) {
        this.send = (message) => { };
        this.onNotification = (name, data, ctx) => { };
        for (const call of this.activeStreamCalls.values()) {
            call.req$.error(new error_1.RpcError(reason));
        }
        this.activeStreamCalls.clear();
    }
    disconnect() {
        this.stop(constants_1.RpcServerError.Disconnect);
    }
    sendError(id, code) {
        const data = this.error.formatCode(code);
        const message = new nominal_1.ResponseErrorMessage(id, data);
        this.send(message);
    }
    execStaticCall(id, name, request, ctx) {
        this.caller.call(name, request, ctx)
            .then(response => {
            this.send(new nominal_1.ResponseCompleteMessage(id, response));
        })
            .catch(error => {
            this.send(new nominal_1.ResponseErrorMessage(id, error));
        });
    }
    createStreamCall(id, name, ctx) {
        const call = this.caller.createCall(name, ctx);
        this.activeStreamCalls.set(id, call);
        subscribeCompleteObserver_1.subscribeCompleteObserver(call.res$, {
            next: (value) => {
                this.send(new nominal_1.ResponseDataMessage(id, value));
            },
            error: (error) => {
                this.send(new nominal_1.ResponseErrorMessage(id, error));
                this.activeStreamCalls.delete(id);
            },
            complete: (value) => {
                this.send(new nominal_1.ResponseCompleteMessage(id, value));
                this.activeStreamCalls.delete(id);
            },
        });
        call.reqUnsubscribe$.subscribe(() => {
            if (this.activeStreamCalls.has(id))
                this.send(new nominal_1.RequestUnsubscribeMessage(id));
        });
        return call;
    }
    onRequestDataMessage(message, ctx) {
        const { id, method, data } = message;
        let call = this.activeStreamCalls.get(id);
        if (!call) {
            if (!method) {
                this.sendError(id, constants_1.RpcServerError.NoMethodSpecified);
                return;
            }
            call = this.createStreamCall(id, method, ctx);
        }
        if (call) {
            if (data !== undefined) {
                call.req$.next(data);
            }
        }
    }
    onRequestCompleteMessage(message, ctx) {
        const { id, method, data } = message;
        const call = this.activeStreamCalls.get(id);
        if (call) {
            const { req$ } = call;
            if (data !== undefined)
                req$.next(data);
            req$.complete();
            return;
        }
        if (!method) {
            this.sendError(id, constants_1.RpcServerError.NoMethodSpecified);
            return;
        }
        if (!this.caller.exists(method)) {
            this.sendError(id, constants_1.RpcServerError.MethodNotFound);
            return;
        }
        const rpcMethod = this.caller.get(method);
        if (rpcMethod.isStreaming) {
            const newCall = this.createStreamCall(id, method, ctx);
            if (newCall) {
                if (data !== undefined) {
                    newCall.req$.next(data);
                }
            }
        }
        else
            this.execStaticCall(id, method, data, ctx);
    }
    onRequestErrorMessage(message, ctx) {
        const { id, method, data } = message;
        const call = this.activeStreamCalls.get(id);
        if (call) {
            return call.req$.error(data);
        }
        if (!method)
            return this.sendError(id, constants_1.RpcServerError.NoMethodSpecified);
        if (!this.caller.exists(method))
            return this.sendError(id, constants_1.RpcServerError.MethodNotFound);
        const rpcMethod = this.caller.get(method);
        if (!rpcMethod.isStreaming)
            return this.sendError(id, constants_1.RpcServerError.ErrorForStaticMethod);
        const streamCall = this.createStreamCall(id, rpcMethod, ctx);
        if (!streamCall)
            return;
        streamCall.req$.error(data);
    }
    onUnsubscribeMessage(message) {
        const { id } = message;
        const call = this.activeStreamCalls.get(id);
        if (!call)
            return;
        this.activeStreamCalls.delete(id);
        call.req$.complete();
    }
    onNotificationMessage(message, ctx) {
        const { method, data } = message;
        if (!method || (method.length > 128))
            throw new error_1.RpcError(constants_1.RpcServerError.InvalidNotificationName);
        this.onNotification(method, data, ctx);
    }
}
exports.RpcServer = RpcServer;
