import { NotificationMessage, SubscribeMessage, DataMessage, CompleteMessage, UnsubscribeMessage, ErrorMessage, } from './messages';
import { from, isObservable, of } from 'rxjs';
import { mergeMap } from 'rxjs/operators';
import { TimedQueue } from '../json-rx/TimedQueue';
import { decodeFullMessages, Encoder } from './codec';
import { assertName, microtask } from '../json-rx/util';
import { isUint8Array } from '../util/isUint8Array';
const ERR_UNKNOWN = new Uint8Array([0]);
const ERR_ID_TAKEN = new Uint8Array([1]);
const ERR_TOO_MANY_SUBSCRIPTIONS = new Uint8Array([2]);
const ERR_INVALID_DATA = new Uint8Array([3]);
export class BinaryRxServer {
    constructor({ send, call, notify, maxActiveSubscriptions = 30, bufferSize = 10, bufferTime = 1, }) {
        this.active = new Map();
        this.encoder = new Encoder();
        this.call = call;
        this.notify = notify;
        this.maxActiveSubscriptions = maxActiveSubscriptions;
        if (bufferTime) {
            const buffer = new TimedQueue();
            buffer.itemLimit = bufferSize;
            buffer.timeLimit = bufferTime;
            buffer.onFlush = (messages) => {
                send(this.encoder.encode(messages));
            };
            this.send = (message) => buffer.push(message);
        }
        else {
            this.send = (message) => {
                send(this.encoder.encode([message]));
            };
        }
    }
    sendError(id, data) {
        this.send(new ErrorMessage(id, data));
    }
    onSubscribe(message, ctx) {
        const { id, method, data } = message;
        assertName(method);
        try {
            const active = this.active.get(id);
            if (active) {
                this.active.delete(id);
                active.unsubscribe();
                this.sendError(id, ERR_ID_TAKEN);
                return;
            }
            if (this.active.size >= this.maxActiveSubscriptions)
                return this.sendError(id, ERR_TOO_MANY_SUBSCRIPTIONS);
            const callResult = this.call(method, data, ctx);
            const observable = (isObservable(callResult)
                ? callResult
                : from(callResult).pipe(mergeMap((value) => (isObservable(value) ? value : of(value)))));
            const ref = { buffer: [] };
            let done = false;
            const subscription = observable.subscribe((data) => {
                if (!isUint8Array(data))
                    return this.sendError(id, ERR_INVALID_DATA);
                if (data === undefined)
                    return;
                ref.buffer.push(data);
                microtask(() => {
                    if (!ref.buffer.length)
                        return;
                    try {
                        for (const payload of ref.buffer)
                            this.send(new DataMessage(id, payload));
                    }
                    finally {
                        ref.buffer = [];
                    }
                });
            }, (error) => {
                done = true;
                this.active.delete(id);
                this.sendError(id, isUint8Array(error) ? error : ERR_UNKNOWN);
            }, () => {
                done = true;
                this.active.delete(id);
                try {
                    if (!ref.buffer.length)
                        return this.send(new CompleteMessage(id, undefined));
                    const last = ref.buffer.length - 1;
                    for (let i = 0; i <= last; i++) {
                        const isLast = i === last;
                        const message = isLast
                            ? new CompleteMessage(id, ref.buffer[i])
                            : new DataMessage(id, ref.buffer[i]);
                        this.send(message);
                    }
                }
                finally {
                    ref.buffer = [];
                }
            });
            if (!done)
                this.active.set(id, subscription);
        }
        catch (error) {
            this.sendError(id, Buffer.isBuffer(error) ? error : ERR_UNKNOWN);
        }
    }
    onUnsubscribe(message) {
        const { id } = message;
        const subscription = this.active.get(id);
        if (!subscription)
            return;
        this.active.delete(id);
        subscription.unsubscribe();
    }
    onNotification(message, ctx) {
        const { method, data } = message;
        assertName(method);
        this.notify(method, data, ctx);
    }
    onArray(arr, ctx) {
        const messages = decodeFullMessages(arr, 0);
        this.onMessages(messages, ctx);
    }
    onMessages(messages, ctx) {
        const length = messages.length;
        for (let i = 0; i < length; i++) {
            const message = messages[i];
            if (message instanceof NotificationMessage)
                this.onNotification(message, ctx);
            else if (message instanceof SubscribeMessage)
                this.onSubscribe(message, ctx);
            else if (message instanceof UnsubscribeMessage)
                this.onUnsubscribe(message);
        }
    }
    stop() {
        this.send = (message) => { };
        for (const sub of this.active.values()) {
            sub.unsubscribe();
        }
    }
}
