import { LogicalTimestamp } from './clock';
import { DeleteOperation } from './operations/DeleteOperation';
import { InsertArrayElementsOperation } from './operations/InsertArrayElementsOperation';
import { InsertStringSubstringOperation } from './operations/InsertStringSubstringOperation';
import { MakeArrayOperation } from './operations/MakeArrayOperation';
import { MakeNumberOperation } from './operations/MakeNumberOperation';
import { MakeObjectOperation } from './operations/MakeObjectOperation';
import { MakeStringOperation } from './operations/MakeStringOperation';
import { SetObjectKeysOperation } from './operations/SetObjectKeysOperation';
import { SetRootOperation } from './operations/SetRootOperation';
import { SetNumberOperation } from './operations/SetNumberOperation';
import { NoopOperation } from './operations/NoopOperation';
import { MakeConstantOperation } from './operations/MakeConstantOperation';
import { MakeValueOperation } from './operations/MakeValueOperation';
import { SetValueOperation } from './operations/SetValueOperation';
import { FALSE_ID, NULL_ID, TRUE_ID, UNDEFINED_ID } from './constants';
import { Patch } from './Patch';
export class PatchBuilder {
    constructor(clock) {
        this.clock = clock;
        this.patch = new Patch();
    }
    obj() {
        this.pad();
        const id = this.clock.tick(1);
        const op = new MakeObjectOperation(id);
        this.patch.ops.push(op);
        return id;
    }
    arr() {
        this.pad();
        const id = this.clock.tick(1);
        const op = new MakeArrayOperation(id);
        this.patch.ops.push(op);
        return id;
    }
    str() {
        this.pad();
        const id = this.clock.tick(1);
        const op = new MakeStringOperation(id);
        this.patch.ops.push(op);
        return id;
    }
    num() {
        this.pad();
        const id = this.clock.tick(1);
        const op = new MakeNumberOperation(id);
        this.patch.ops.push(op);
        return id;
    }
    const(value) {
        this.pad();
        const id = this.clock.tick(1);
        const op = new MakeConstantOperation(id, value);
        this.patch.ops.push(op);
        return id;
    }
    val(value) {
        this.pad();
        const id = this.clock.tick(1);
        const op = new MakeValueOperation(id, value);
        this.patch.ops.push(op);
        return id;
    }
    root(value) {
        this.pad();
        const id = this.clock.tick(1);
        const op = new SetRootOperation(id, value);
        this.patch.ops.push(op);
        return id;
    }
    setKeys(obj, tuples) {
        this.pad();
        if (!tuples.length)
            throw new Error('EMPTY_TUPLES');
        const id = this.clock.tick(1);
        const op = new SetObjectKeysOperation(id, obj, tuples);
        const span = op.span();
        if (span > 1)
            this.clock.tick(span - 1);
        this.patch.ops.push(op);
        return id;
    }
    setNum(obj, value) {
        this.pad();
        const id = this.clock.tick(1);
        const op = new SetNumberOperation(id, obj, value);
        this.patch.ops.push(op);
        return id;
    }
    setVal(obj, value) {
        this.pad();
        const id = this.clock.tick(1);
        const op = new SetValueOperation(id, obj, value);
        this.patch.ops.push(op);
        return id;
    }
    insStr(obj, after, substring) {
        this.pad();
        if (!substring.length)
            throw new Error('EMPTY_STRING');
        const id = this.clock.tick(1);
        const op = new InsertStringSubstringOperation(id, obj, after, substring);
        const span = op.span();
        if (span > 1)
            this.clock.tick(span - 1);
        this.patch.ops.push(op);
        return id;
    }
    insArr(arr, after, elements) {
        this.pad();
        const id = this.clock.tick(1);
        const op = new InsertArrayElementsOperation(id, arr, after, elements);
        const span = op.span();
        if (span > 1)
            this.clock.tick(span - 1);
        this.patch.ops.push(op);
        return id;
    }
    del(obj, start, span) {
        this.pad();
        const id = this.clock.tick(span);
        const op = new DeleteOperation(id, obj, start, span);
        this.patch.ops.push(op);
        return id;
    }
    noop(span) {
        this.pad();
        const id = this.clock.tick(span);
        const op = new NoopOperation(id, span);
        this.patch.ops.push(op);
        return id;
    }
    jsonObj(json) {
        const obj = this.obj();
        const keys = Object.keys(json);
        if (keys.length) {
            const tuples = [];
            for (const k of keys)
                tuples.push([k, this.json(json[k])]);
            this.setKeys(obj, tuples);
        }
        return obj;
    }
    jsonArr(json) {
        const arr = this.arr();
        if (json.length) {
            const values = [];
            for (const el of json)
                values.push(this.json(el));
            this.insArr(arr, arr, values);
        }
        return arr;
    }
    jsonStr(json) {
        const str = this.str();
        if (json)
            this.insStr(str, str, json);
        return str;
    }
    jsonVal(json) {
        return this.val(json);
    }
    json(json) {
        switch (json) {
            case null:
                return NULL_ID;
            case true:
                return TRUE_ID;
            case false:
                return FALSE_ID;
        }
        if (Array.isArray(json))
            return this.jsonArr(json);
        switch (typeof json) {
            case 'object':
                return this.jsonObj(json);
            case 'string':
                return this.jsonStr(json);
            case 'number':
                return this.jsonVal(json);
        }
        return UNDEFINED_ID;
    }
    pad() {
        const nextTime = this.patch.nextTime();
        if (!nextTime)
            return;
        const drift = this.clock.time - nextTime;
        if (drift > 0) {
            const id = new LogicalTimestamp(this.clock.getSessionId(), nextTime);
            const padding = new NoopOperation(id, drift);
            this.patch.ops.push(padding);
        }
    }
}
