import { FALSE, NULL, TRUE, UNDEFINED } from '../constants';
import { LogicalNodeIndex, ServerNodeIndex } from './nodes';
import { randomSessionId } from './util';
import { SetRootOperation } from '../../json-crdt-patch/operations/SetRootOperation';
import { LogicalVectorClock, ServerVectorClock } from '../../json-crdt-patch/clock';
import { DocRootType } from '../types/lww-doc-root/DocRootType';
import { ObjectType } from '../types/lww-object/ObjectType';
import { ArrayType } from '../types/rga-array/ArrayType';
import { StringType } from '../types/rga-string/StringType';
import { MakeObjectOperation } from '../../json-crdt-patch/operations/MakeObjectOperation';
import { SetObjectKeysOperation } from '../../json-crdt-patch/operations/SetObjectKeysOperation';
import { MakeArrayOperation } from '../../json-crdt-patch/operations/MakeArrayOperation';
import { InsertArrayElementsOperation } from '../../json-crdt-patch/operations/InsertArrayElementsOperation';
import { ORIGIN } from '../../json-crdt-patch/constants';
import { DeleteOperation } from '../../json-crdt-patch/operations/DeleteOperation';
import { MakeStringOperation } from '../../json-crdt-patch/operations/MakeStringOperation';
import { InsertStringSubstringOperation } from '../../json-crdt-patch/operations/InsertStringSubstringOperation';
import { ModelApi } from './api/ModelApi';
import { MakeValueOperation } from '../../json-crdt-patch/operations/MakeValueOperation';
import { ValueType } from '../types/lww-value/ValueType';
import { SetValueOperation } from '../../json-crdt-patch/operations/SetValueOperation';
export class Model {
    constructor(clock, nodes) {
        this.root = new DocRootType(this, ORIGIN, null);
        this.clock = clock;
        this.nodes = nodes;
        this.api = new ModelApi(this);
    }
    static withLogicalClock(clock) {
        clock = clock || new LogicalVectorClock(randomSessionId(), 0);
        const nodes = new LogicalNodeIndex();
        return new Model(clock, nodes);
    }
    static withServerClock(time) {
        const clock = new ServerVectorClock(time || 0);
        const nodes = new ServerNodeIndex();
        return new Model(clock, nodes);
    }
    node(id) {
        if (id.getSessionId() === 0) {
            switch (id.time) {
                case 1: return NULL;
                case 2: return TRUE;
                case 3: return FALSE;
                case 4: return UNDEFINED;
            }
        }
        return this.nodes.get(id);
    }
    applyPatch(patch) {
        const ops = patch.ops;
        const { length } = ops;
        for (let i = 0; i < length; i++)
            this.applyOperation(ops[i]);
    }
    applyOperation(op) {
        this.clock.observe(op.id, op.span());
        if (op instanceof MakeObjectOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ObjectType(this, op.id));
        }
        else if (op instanceof MakeArrayOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ArrayType(this, op.id));
        }
        else if (op instanceof MakeStringOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new StringType(this, op.id));
        }
        else if (op instanceof MakeValueOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ValueType(op.id, op.id, op.value));
        }
        else if (op instanceof SetRootOperation) {
            const oldValue = this.root.insert(op);
            if (oldValue)
                this.deleteNodeTree(oldValue);
        }
        else if (op instanceof SetObjectKeysOperation) {
            const obj = this.nodes.get(op.object);
            if (obj instanceof ObjectType)
                obj.insert(op);
        }
        else if (op instanceof SetValueOperation) {
            const obj = this.nodes.get(op.obj);
            if (obj instanceof ValueType)
                obj.insert(op);
        }
        else if (op instanceof InsertArrayElementsOperation) {
            const arr = this.nodes.get(op.arr);
            if (arr instanceof ArrayType)
                arr.insert(op);
        }
        else if (op instanceof InsertStringSubstringOperation) {
            const arr = this.nodes.get(op.obj);
            if (arr instanceof StringType)
                arr.onInsert(op);
        }
        else if (op instanceof DeleteOperation) {
            const node = this.nodes.get(op.obj);
            if (node instanceof ArrayType)
                node.delete(op);
            else if (node instanceof StringType)
                node.onDelete(op);
        }
    }
    deleteNodeTree(value) {
        const isSystemNode = value.getSessionId() < 1;
        if (isSystemNode)
            return;
        const node = this.nodes.get(value);
        if (!node)
            return;
        for (const child of node.children())
            this.deleteNodeTree(child);
        this.nodes.delete(value);
    }
    clone() {
        const model = this.clock instanceof LogicalVectorClock
            ? Model.withLogicalClock(this.clock.clone())
            : Model.withServerClock(this.clock.time);
        model.root = this.root.clone(model);
        return model;
    }
    fork(sessionId = randomSessionId()) {
        const model = this.clock instanceof LogicalVectorClock
            ? Model.withLogicalClock(this.clock.fork(sessionId))
            : Model.withServerClock(this.clock.time);
        model.root = this.root.clone(model);
        return model;
    }
    toJson() {
        return this.root.toJson();
    }
    toString(tab) {
        return this.root.toString(tab);
    }
}
