import { firstValueFrom, isObservable, Observable, Subject } from 'rxjs';
import { NotificationMessage, RequestCompleteMessage, RequestDataMessage, RequestErrorMessage, ResponseCompleteMessage, ResponseDataMessage, ResponseErrorMessage, ResponseUnsubscribeMessage } from '../messages/nominal';
import { subscribeCompleteObserver } from '../util/subscribeCompleteObserver';
import { TimedQueue } from '../util/TimedQueue';
export class RpcClient {
    constructor({ send, bufferSize = 100, bufferTime = 10 }) {
        this.id = 1;
        this.calls = new Map();
        this.buffer = new TimedQueue();
        this.buffer.itemLimit = bufferSize;
        this.buffer.timeLimit = bufferTime;
        this.buffer.onFlush = send;
    }
    getInflightCallCount() {
        return this.calls.size;
    }
    onMessages(messages) {
        const length = messages.length;
        for (let i = 0; i < length; i++)
            this.onMessage(messages[i]);
    }
    onMessage(message) {
        if (message instanceof ResponseCompleteMessage)
            return this.onResponseComplete(message);
        else if (message instanceof ResponseDataMessage)
            return this.onResponseData(message);
        else if (message instanceof ResponseErrorMessage)
            return this.onResponseError(message);
        return this.onRequestUnsubscribe(message);
    }
    onResponseComplete(message) {
        const { id, data } = message;
        const call = this.calls.get(id);
        if (!call)
            return;
        call.resFinalized = true;
        if (data !== undefined)
            call.res$.next(data);
        call.res$.complete();
    }
    onResponseData(message) {
        const { id, data } = message;
        const call = this.calls.get(id);
        if (!call)
            return;
        call.res$.next(data);
    }
    onResponseError(message) {
        const { id, data } = message;
        const call = this.calls.get(id);
        if (!call)
            return;
        call.resFinalized = true;
        call.res$.error(data);
    }
    onRequestUnsubscribe(message) {
        const { id } = message;
        const call = this.calls.get(id);
        if (!call)
            return;
        call.req$.complete();
    }
    call$(method, data) {
        const id = this.id++;
        if (this.id >= 0xffff)
            this.id = 1;
        if (this.calls.has(id))
            return this.call$(method, data);
        const req$ = new Subject();
        const res$ = new Subject();
        let finalizedStreams = 0;
        const cleanup = () => {
            finalizedStreams++;
            if (finalizedStreams === 2)
                this.calls.delete(id);
        };
        res$.subscribe({ error: cleanup, complete: cleanup });
        const entry = { req$, res$ };
        this.calls.set(id, entry);
        if (isObservable(data)) {
            let firstMessageSent = false;
            subscribeCompleteObserver(req$, {
                next: (value) => {
                    const messageMethod = firstMessageSent ? '' : method;
                    firstMessageSent = true;
                    const message = new RequestDataMessage(id, messageMethod, value);
                    this.buffer.push(message);
                },
                error: (error) => {
                    cleanup();
                    const messageMethod = firstMessageSent ? '' : method;
                    const message = new RequestErrorMessage(id, messageMethod, error);
                    this.buffer.push(message);
                },
                complete: (value) => {
                    cleanup();
                    const messageMethod = firstMessageSent ? '' : method;
                    const message = new RequestCompleteMessage(id, messageMethod, value);
                    this.buffer.push(message);
                },
            });
            data.subscribe(req$);
        }
        else {
            this.buffer.push(new RequestCompleteMessage(id, method, data));
            req$.complete();
            cleanup();
        }
        return new Observable((observer) => {
            res$.subscribe(observer);
            return () => {
                if (!entry.resFinalized)
                    this.buffer.push(new ResponseUnsubscribeMessage(id));
                res$.complete();
            };
        });
    }
    async call(method, request) {
        return await firstValueFrom(this.call$(method, request));
    }
    notify(method, data) {
        this.buffer.push(new NotificationMessage(method, data));
    }
    stop(reason = 'STOP') {
        this.buffer.onFlush = (message) => { };
        for (const call of this.calls.values()) {
            call.req$.error(new Error(reason));
            call.req$.error(new Error(reason));
        }
        this.calls.clear();
    }
    disconnect() {
        this.stop('DISCONNECT');
    }
}
