import { Encoder as EncoderJson, Decoder as DecoderJson } from '../../../common/codec/compact-json';
import { Encoder as EncoderMsgPack, Decoder as DecoderMsgPack } from '../../../common/codec/compact-msgpack';
import { NotificationMessage } from '../../../common/messages/nominal/NotificationMessage';
import { createConnectionContext } from '../context';
export const enableWsCompactReactiveRpcApi = (params) => {
    const encoderJson = new EncoderJson();
    const decoderJson = new DecoderJson();
    const encoderMsgPack = new EncoderMsgPack();
    const decoderMsgPack = new DecoderMsgPack();
    const invalidPayloadNotification = new NotificationMessage('.err', 'CODING');
    const invalidPayloadJson = encoderJson.encode([invalidPayloadNotification]);
    const invalidPayloadMsgPack = encoderMsgPack.encode([invalidPayloadNotification]);
    const { route = '/rpc/compact', uws, createRpcServer, onNotification, createContext = createConnectionContext, compression, idleTimeout = 0, maxBackpressure = 3145728, maxPayloadLength = 1048576, } = params;
    uws.ws(route, {
        idleTimeout,
        compression,
        maxBackpressure,
        maxPayloadLength,
        upgrade: (res, req, context) => {
            const secWebSocketKey = req.getHeader('sec-websocket-key');
            const secWebSocketProtocol = req.getHeader('sec-websocket-protocol');
            const secWebSocketExtensions = req.getHeader('sec-websocket-extensions');
            const ctx = createContext(req, res);
            const isBinary = secWebSocketProtocol.indexOf('MsgPack') > -1;
            res.upgrade({ ctx, isBinary }, secWebSocketKey, secWebSocketProtocol, secWebSocketExtensions, context);
        },
        open: (ws) => {
            const rpc = createRpcServer({
                send: (messages) => {
                    if (ws.getBufferedAmount() > maxBackpressure)
                        return;
                    const { isBinary } = ws;
                    const encoded = isBinary ? encoderMsgPack.encode(messages) : encoderJson.encode(messages);
                    ws.send(encoded, isBinary);
                },
            });
            if (onNotification) {
                rpc.onNotification = (name, data, ctx) => {
                    onNotification(ws, name, data, ctx);
                };
            }
            ws.rpc = rpc;
        },
        message: (ws, buf, isBinary) => {
            const { ctx, rpc } = ws;
            let messages;
            try {
                messages = isBinary
                    ? decoderMsgPack.decode(new Uint8Array(buf))
                    : decoderJson.decode(Buffer.from(buf).toString('utf8'));
            }
            catch (error) {
                if (ws.getBufferedAmount() > maxBackpressure)
                    return;
                ws.send(isBinary ? invalidPayloadMsgPack : invalidPayloadJson, isBinary);
                return;
            }
            if (messages instanceof Array)
                rpc.onMessages(messages, ctx);
            else
                rpc.onMessage(messages, ctx);
        },
        close: (ws, code, message) => {
            ws.rpc.stop();
        },
    });
};
