"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Model = void 0;
var tslib_1 = require("tslib");
var constants_1 = require("../constants");
var nodes_1 = require("./nodes");
var util_1 = require("./util");
var SetRootOperation_1 = require("../../json-crdt-patch/operations/SetRootOperation");
var clock_1 = require("../../json-crdt-patch/clock");
var DocRootType_1 = require("../types/lww-doc-root/DocRootType");
var ObjectType_1 = require("../types/lww-object/ObjectType");
var ArrayType_1 = require("../types/rga-array/ArrayType");
var StringType_1 = require("../types/rga-string/StringType");
var MakeObjectOperation_1 = require("../../json-crdt-patch/operations/MakeObjectOperation");
var SetObjectKeysOperation_1 = require("../../json-crdt-patch/operations/SetObjectKeysOperation");
var MakeArrayOperation_1 = require("../../json-crdt-patch/operations/MakeArrayOperation");
var InsertArrayElementsOperation_1 = require("../../json-crdt-patch/operations/InsertArrayElementsOperation");
var constants_2 = require("../../json-crdt-patch/constants");
var DeleteOperation_1 = require("../../json-crdt-patch/operations/DeleteOperation");
var MakeStringOperation_1 = require("../../json-crdt-patch/operations/MakeStringOperation");
var InsertStringSubstringOperation_1 = require("../../json-crdt-patch/operations/InsertStringSubstringOperation");
var ModelApi_1 = require("./api/ModelApi");
var MakeValueOperation_1 = require("../../json-crdt-patch/operations/MakeValueOperation");
var ValueType_1 = require("../types/lww-value/ValueType");
var SetValueOperation_1 = require("../../json-crdt-patch/operations/SetValueOperation");
var Model = (function () {
    function Model(clock, nodes) {
        this.root = new DocRootType_1.DocRootType(this, constants_2.ORIGIN, null);
        this.clock = clock;
        this.nodes = nodes;
        this.api = new ModelApi_1.ModelApi(this);
    }
    Model.withLogicalClock = function (clock) {
        clock = clock || new clock_1.LogicalVectorClock(util_1.randomSessionId(), 0);
        var nodes = new nodes_1.LogicalNodeIndex();
        return new Model(clock, nodes);
    };
    Model.withServerClock = function (time) {
        var clock = new clock_1.ServerVectorClock(time || 0);
        var nodes = new nodes_1.ServerNodeIndex();
        return new Model(clock, nodes);
    };
    Model.prototype.node = function (id) {
        if (id.getSessionId() === 0) {
            switch (id.time) {
                case 1: return constants_1.NULL;
                case 2: return constants_1.TRUE;
                case 3: return constants_1.FALSE;
                case 4: return constants_1.UNDEFINED;
            }
        }
        return this.nodes.get(id);
    };
    Model.prototype.applyPatch = function (patch) {
        var ops = patch.ops;
        var length = ops.length;
        for (var i = 0; i < length; i++)
            this.applyOperation(ops[i]);
    };
    Model.prototype.applyOperation = function (op) {
        this.clock.observe(op.id, op.span());
        if (op instanceof MakeObjectOperation_1.MakeObjectOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ObjectType_1.ObjectType(this, op.id));
        }
        else if (op instanceof MakeArrayOperation_1.MakeArrayOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ArrayType_1.ArrayType(this, op.id));
        }
        else if (op instanceof MakeStringOperation_1.MakeStringOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new StringType_1.StringType(this, op.id));
        }
        else if (op instanceof MakeValueOperation_1.MakeValueOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ValueType_1.ValueType(op.id, op.id, op.value));
        }
        else if (op instanceof SetRootOperation_1.SetRootOperation) {
            var oldValue = this.root.insert(op);
            if (oldValue)
                this.deleteNodeTree(oldValue);
        }
        else if (op instanceof SetObjectKeysOperation_1.SetObjectKeysOperation) {
            var obj = this.nodes.get(op.object);
            if (obj instanceof ObjectType_1.ObjectType)
                obj.insert(op);
        }
        else if (op instanceof SetValueOperation_1.SetValueOperation) {
            var obj = this.nodes.get(op.obj);
            if (obj instanceof ValueType_1.ValueType)
                obj.insert(op);
        }
        else if (op instanceof InsertArrayElementsOperation_1.InsertArrayElementsOperation) {
            var arr = this.nodes.get(op.arr);
            if (arr instanceof ArrayType_1.ArrayType)
                arr.insert(op);
        }
        else if (op instanceof InsertStringSubstringOperation_1.InsertStringSubstringOperation) {
            var arr = this.nodes.get(op.obj);
            if (arr instanceof StringType_1.StringType)
                arr.onInsert(op);
        }
        else if (op instanceof DeleteOperation_1.DeleteOperation) {
            var node = this.nodes.get(op.obj);
            if (node instanceof ArrayType_1.ArrayType)
                node.delete(op);
            else if (node instanceof StringType_1.StringType)
                node.onDelete(op);
        }
    };
    Model.prototype.deleteNodeTree = function (value) {
        var e_1, _a;
        var isSystemNode = value.getSessionId() < 1;
        if (isSystemNode)
            return;
        var node = this.nodes.get(value);
        if (!node)
            return;
        try {
            for (var _b = tslib_1.__values(node.children()), _c = _b.next(); !_c.done; _c = _b.next()) {
                var child = _c.value;
                this.deleteNodeTree(child);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_1) throw e_1.error; }
        }
        this.nodes.delete(value);
    };
    Model.prototype.clone = function () {
        var model = this.clock instanceof clock_1.LogicalVectorClock
            ? Model.withLogicalClock(this.clock.clone())
            : Model.withServerClock(this.clock.time);
        model.root = this.root.clone(model);
        return model;
    };
    Model.prototype.fork = function (sessionId) {
        if (sessionId === void 0) { sessionId = util_1.randomSessionId(); }
        var model = this.clock instanceof clock_1.LogicalVectorClock
            ? Model.withLogicalClock(this.clock.fork(sessionId))
            : Model.withServerClock(this.clock.time);
        model.root = this.root.clone(model);
        return model;
    };
    Model.prototype.toJson = function () {
        return this.root.toJson();
    };
    Model.prototype.toString = function (tab) {
        return this.root.toString(tab);
    };
    return Model;
}());
exports.Model = Model;
