"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ModelApi = void 0;
var tslib_1 = require("tslib");
var StringType_1 = require("../../types/rga-string/StringType");
var PatchBuilder_1 = require("../../../json-crdt-patch/PatchBuilder");
var Patch_1 = require("../../../json-crdt-patch/Patch");
var NoopOperation_1 = require("../../../json-crdt-patch/operations/NoopOperation");
var clock_1 = require("../../../json-crdt-patch/clock");
var StringApi_1 = require("./StringApi");
var ArrayType_1 = require("../../types/rga-array/ArrayType");
var ObjectType_1 = require("../../types/lww-object/ObjectType");
var constants_1 = require("../../../json-crdt-patch/constants");
var ValueType_1 = require("../../types/lww-value/ValueType");
var constants_2 = require("../../constants");
var ModelApi = (function () {
    function ModelApi(doc) {
        this.doc = doc;
        this.patches = [];
        this.builder = new PatchBuilder_1.PatchBuilder(doc.clock);
    }
    ModelApi.prototype.commit = function () {
        var patch = this.builder.patch;
        this.builder = new PatchBuilder_1.PatchBuilder(this.doc.clock);
        if (patch.ops.length) {
            this.patches.push(patch);
            this.doc.applyPatch(patch);
        }
        return patch;
    };
    ModelApi.prototype.flush = function () {
        var patches = this.patches;
        this.patches = [];
        return patches;
    };
    ModelApi.prototype.flushPatch = function () {
        var _a, _b;
        var patches = this.flush();
        var length = patches.length;
        var result = new Patch_1.Patch();
        var prev = null;
        for (var i = 0; i < length; i++) {
            var patch = patches[i];
            if (!patch.ops)
                continue;
            if (!prev)
                (_a = result.ops).push.apply(_a, tslib_1.__spreadArray([], tslib_1.__read(patch.ops)));
            else {
                var id = patch.getId();
                var nextTime = prev.nextTime();
                var gap = id.time - nextTime;
                if (gap > 0)
                    result.ops.push(new NoopOperation_1.NoopOperation(new clock_1.LogicalTimestamp(id.getSessionId(), nextTime), gap));
                (_b = result.ops).push.apply(_b, tslib_1.__spreadArray([], tslib_1.__read(patch.ops)));
            }
            prev = patch;
        }
        return result;
    };
    ModelApi.prototype.find = function (steps) {
        var doc = this.doc;
        var id = doc.root.toValue();
        var node = doc.node(id) || constants_2.NULL;
        var length = steps.length;
        if (!length)
            return node;
        var i = 0;
        while (i < length) {
            var step = steps[i++];
            if (node instanceof ObjectType_1.ObjectType) {
                var id_1 = node.get(String(step));
                if (!id_1)
                    return constants_2.UNDEFINED;
                var nextNode = doc.node(id_1);
                if (!nextNode)
                    return constants_2.UNDEFINED;
                node = nextNode;
            }
            else if (node instanceof ArrayType_1.ArrayType) {
                var id_2 = node.findValue(Number(step));
                var nextNode = doc.node(id_2);
                if (!nextNode)
                    return constants_2.UNDEFINED;
                node = nextNode;
                continue;
            }
        }
        return node;
    };
    ModelApi.prototype.patch = function (cb) {
        this.commit();
        cb(this);
        this.commit();
    };
    ModelApi.prototype.root = function (json) {
        var value = this.builder.json(json);
        this.builder.root(value);
        return this;
    };
    ModelApi.prototype.asStr = function (path) {
        var obj = this.find(path);
        if (obj instanceof StringType_1.StringType)
            return obj;
        throw new Error('NOT_STR');
    };
    ModelApi.prototype.str = function (path) {
        var obj = this.asStr(path);
        return new StringApi_1.StringApi(this, obj);
    };
    ModelApi.prototype.strObjIns = function (obj, index, substr) {
        var after = !index ? obj.id : obj.findId(index - 1);
        this.builder.insStr(obj.id, after, substr);
    };
    ModelApi.prototype.strIns = function (path, index, substr) {
        var obj = this.asStr(path);
        this.strObjIns(obj, index, substr);
        return this;
    };
    ModelApi.prototype.strObjDel = function (obj, index, length) {
        var e_1, _a;
        var spans = obj.findIdSpan(index, length);
        try {
            for (var spans_1 = tslib_1.__values(spans), spans_1_1 = spans_1.next(); !spans_1_1.done; spans_1_1 = spans_1.next()) {
                var ts = spans_1_1.value;
                this.builder.del(obj.id, ts, ts.span);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (spans_1_1 && !spans_1_1.done && (_a = spans_1.return)) _a.call(spans_1);
            }
            finally { if (e_1) throw e_1.error; }
        }
    };
    ModelApi.prototype.strDel = function (path, index, length) {
        var obj = this.asStr(path);
        this.strObjDel(obj, index, length);
        return this;
    };
    ModelApi.prototype.asVal = function (path) {
        var obj = this.find(path);
        if (obj instanceof ValueType_1.ValueType)
            return obj;
        throw new Error('NOT_VAL');
    };
    ModelApi.prototype.valSet = function (path, value) {
        var id = this.asVal(path).id;
        this.builder.setVal(id, value);
        return this;
    };
    ModelApi.prototype.asArr = function (path) {
        var obj = this.find(path);
        if (obj instanceof ArrayType_1.ArrayType)
            return obj;
        throw new Error('NOT_ARR');
    };
    ModelApi.prototype.arrIns = function (path, index, values) {
        var builder = this.builder;
        var obj = this.asArr(path);
        var after = !index ? obj.id : obj.findId(index - 1);
        var valueIds = [];
        for (var i = 0; i < values.length; i++)
            valueIds.push(builder.json(values[i]));
        builder.insArr(obj.id, after, valueIds);
        return this;
    };
    ModelApi.prototype.arrDel = function (path, index, length) {
        var e_2, _a;
        var obj = this.asArr(path);
        var spans = obj.findIdSpans(index, length);
        try {
            for (var spans_2 = tslib_1.__values(spans), spans_2_1 = spans_2.next(); !spans_2_1.done; spans_2_1 = spans_2.next()) {
                var ts = spans_2_1.value;
                this.builder.del(obj.id, ts, ts.span);
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (spans_2_1 && !spans_2_1.done && (_a = spans_2.return)) _a.call(spans_2);
            }
            finally { if (e_2) throw e_2.error; }
        }
        return this;
    };
    ModelApi.prototype.asObj = function (path) {
        var obj = this.find(path);
        if (obj instanceof ObjectType_1.ObjectType)
            return obj;
        throw new Error('NOT_OBJ');
    };
    ModelApi.prototype.objSet = function (path, entries) {
        var obj = this.asObj(path);
        var builder = this.builder;
        builder.setKeys(obj.id, Object.entries(entries).map(function (_a) {
            var _b = tslib_1.__read(_a, 2), key = _b[0], json = _b[1];
            return [key, builder.json(json)];
        }));
        return this;
    };
    ModelApi.prototype.objDel = function (path, keys) {
        var obj = this.asObj(path);
        var builder = this.builder;
        builder.setKeys(obj.id, keys.map(function (key) { return [key, constants_1.UNDEFINED_ID]; }));
        return this;
    };
    return ModelApi;
}());
exports.ModelApi = ModelApi;
