"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Encoder = void 0;
var isSafeInteger = Number.isSafeInteger;
var EMPTY_UINT8 = new Uint8Array([]);
var EMPTY_VIEW = new DataView(EMPTY_UINT8.buffer);
var Encoder = (function () {
    function Encoder() {
        this.uint8 = EMPTY_UINT8;
        this.view = EMPTY_VIEW;
        this.offset = 0;
    }
    Encoder.prototype.grow = function (size) {
        var newUint8 = new Uint8Array(size);
        newUint8.set(this.uint8);
        this.uint8 = newUint8;
        this.view = new DataView(newUint8.buffer);
    };
    Encoder.prototype.ensureCapacity = function (capacity) {
        var size = this.offset + capacity;
        if (this.uint8.byteLength < size)
            this.grow(Math.max(size, this.uint8.byteLength * 4));
    };
    Encoder.prototype.reset = function () {
        this.uint8 = new Uint8Array(1024);
        this.view = new DataView(this.uint8.buffer);
        this.offset = 0;
    };
    Encoder.prototype.flush = function () {
        return this.uint8.subarray(0, this.offset);
    };
    Encoder.prototype.encode = function (json) {
        this.reset();
        this.encodeAny(json);
        return this.flush();
    };
    Encoder.prototype.encodeAny = function (json) {
        switch (json) {
            case null:
                return this.u8(0xc0);
            case false:
                return this.u8(0xc2);
            case true:
                return this.u8(0xc3);
        }
        if (json instanceof Array)
            return this.encodeArray(json);
        switch (typeof json) {
            case 'number':
                return this.encodeNumber(json);
            case 'string':
                return this.encodeString(json);
            case 'object':
                return this.encodeObject(json);
        }
    };
    Encoder.prototype.encodeNumber = function (num) {
        if (isSafeInteger(num)) {
            if (num >= 0) {
                if (num <= 127)
                    return this.u8(num);
                if (num <= 0xff)
                    return this.u16((0xcc << 8) | num);
                else if (num <= 0xffff) {
                    this.ensureCapacity(3);
                    this.uint8[this.offset++] = 0xcd;
                    this.uint8[this.offset++] = num >>> 8;
                    this.uint8[this.offset++] = num & 0xff;
                    return;
                }
                else if (num <= 0xffffffff) {
                    this.ensureCapacity(5);
                    this.uint8[this.offset++] = 0xce;
                    this.view.setUint32(this.offset, num);
                    this.offset += 4;
                    return;
                }
            }
            else {
                if (num >= -32)
                    return this.u8(224 | (num + 0x20));
                else if (num > -0x7fff) {
                    this.ensureCapacity(3);
                    this.uint8[this.offset++] = 0xd1;
                    this.view.setInt16(this.offset, num);
                    this.offset += 2;
                    return;
                }
                else if (num > -0x7fffffff) {
                    this.ensureCapacity(5);
                    this.uint8[this.offset++] = 0xd2;
                    this.view.setInt32(this.offset, num);
                    this.offset += 4;
                    return;
                }
            }
        }
        this.ensureCapacity(9);
        this.uint8[this.offset++] = 0xcb;
        this.view.setFloat64(this.offset, num);
        this.offset += 8;
    };
    Encoder.prototype.encodeString = function (str) {
        var length = str.length;
        var maxSize = length * 4;
        this.ensureCapacity(5 + maxSize);
        var output = this.uint8;
        var lengthOffset = this.offset;
        if (maxSize <= 31)
            this.offset++;
        else if (maxSize <= 0xff) {
            output[this.offset++] = 0xd9;
            lengthOffset = this.offset;
            this.offset++;
        }
        else if (maxSize <= 0xffff) {
            output[this.offset++] = 0xda;
            lengthOffset = this.offset;
            this.offset += 2;
        }
        else {
            output[this.offset++] = 0xdb;
            lengthOffset = this.offset;
            this.offset += 4;
        }
        var offset = this.offset;
        var pos = 0;
        while (pos < length) {
            var value = str.charCodeAt(pos++);
            if ((value & 0xffffff80) === 0) {
                output[offset++] = value;
                continue;
            }
            else if ((value & 0xfffff800) === 0) {
                output[offset++] = ((value >> 6) & 0x1f) | 0xc0;
            }
            else {
                if (value >= 0xd800 && value <= 0xdbff) {
                    if (pos < length) {
                        var extra = str.charCodeAt(pos);
                        if ((extra & 0xfc00) === 0xdc00) {
                            ++pos;
                            value = ((value & 0x3ff) << 10) + (extra & 0x3ff) + 0x10000;
                        }
                    }
                }
                if ((value & 0xffff0000) === 0) {
                    output[offset++] = ((value >> 12) & 0x0f) | 0xe0;
                    output[offset++] = ((value >> 6) & 0x3f) | 0x80;
                }
                else {
                    output[offset++] = ((value >> 18) & 0x07) | 0xf0;
                    output[offset++] = ((value >> 12) & 0x3f) | 0x80;
                    output[offset++] = ((value >> 6) & 0x3f) | 0x80;
                }
            }
            output[offset++] = (value & 0x3f) | 0x80;
        }
        this.offset = offset;
        if (maxSize <= 31)
            this.view.setUint8(lengthOffset, 160 | (offset - lengthOffset - 1));
        else if (maxSize <= 0xff)
            this.view.setUint8(lengthOffset, offset - lengthOffset - 1);
        else if (maxSize <= 0xffff)
            this.view.setUint16(lengthOffset, offset - lengthOffset - 2);
        else
            this.view.setUint32(lengthOffset, offset - lengthOffset - 4);
    };
    Encoder.prototype.encodeArray = function (arr) {
        var length = arr.length;
        if (length <= 15)
            this.u8(144 | length);
        else if (length <= 0xffff) {
            this.u8(0xdc);
            this.u16(length);
        }
        else if (length <= 0xffffffff) {
            this.u8(0xdd);
            this.u32(length);
        }
        else
            return;
        for (var i = 0; i < length; i++)
            this.encodeAny(arr[i]);
    };
    Encoder.prototype.encodeObject = function (obj) {
        var keys = Object.keys(obj);
        var length = keys.length;
        if (length <= 15)
            this.u8(128 | length);
        else if (length <= 0xffff) {
            this.u8(0xde);
            this.u16(length);
        }
        else if (length <= 0xffffffff) {
            this.u8(0xdf);
            this.u32(length);
        }
        else
            return;
        for (var i = 0; i < length; i++) {
            var key = keys[i];
            this.encodeString(key);
            this.encodeAny(obj[key]);
        }
    };
    Encoder.prototype.u8 = function (char) {
        this.ensureCapacity(1);
        this.view.setUint8(this.offset++, char);
    };
    Encoder.prototype.u16 = function (word) {
        this.ensureCapacity(2);
        this.view.setUint16(this.offset, word);
        this.offset += 2;
    };
    Encoder.prototype.u32 = function (dword) {
        this.ensureCapacity(4);
        this.view.setUint32(this.offset, dword);
        this.offset += 4;
    };
    return Encoder;
}());
exports.Encoder = Encoder;
