"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Encoder = void 0;
var tslib_1 = require("tslib");
var encoder_1 = require("../../util/encoder");
var isSafeInteger = Number.isSafeInteger;
var Encoder = (function (_super) {
    tslib_1.__extends(Encoder, _super);
    function Encoder() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    Encoder.prototype.encode = function (json) {
        this.reset();
        this.encodeAny(json);
        return this.flush();
    };
    Encoder.prototype.encodeAny = function (json) {
        switch (json) {
            case null:
                return this.u8(0xc0);
            case false:
                return this.u8(0xc2);
            case true:
                return this.u8(0xc3);
        }
        if (json instanceof Array)
            return this.encodeArray(json);
        switch (typeof json) {
            case 'number':
                return this.encodeNumber(json);
            case 'string':
                return this.encodeString(json);
            case 'object':
                return this.encodeObject(json);
        }
    };
    Encoder.prototype.encodeNumber = function (num) {
        if (isSafeInteger(num)) {
            if (num >= 0) {
                if (num <= 127)
                    return this.u8(num);
                if (num <= 0xff)
                    return this.u16((0xcc << 8) | num);
                else if (num <= 0xffff) {
                    this.ensureCapacity(3);
                    this.uint8[this.offset++] = 0xcd;
                    this.uint8[this.offset++] = num >>> 8;
                    this.uint8[this.offset++] = num & 0xff;
                    return;
                }
                else if (num <= 0xffffffff) {
                    this.ensureCapacity(5);
                    this.uint8[this.offset++] = 0xce;
                    this.view.setUint32(this.offset, num);
                    this.offset += 4;
                    return;
                }
            }
            else {
                if (num >= -32)
                    return this.u8(224 | (num + 0x20));
                else if (num > -0x7fff) {
                    this.ensureCapacity(3);
                    this.uint8[this.offset++] = 0xd1;
                    this.view.setInt16(this.offset, num);
                    this.offset += 2;
                    return;
                }
                else if (num > -0x7fffffff) {
                    this.ensureCapacity(5);
                    this.uint8[this.offset++] = 0xd2;
                    this.view.setInt32(this.offset, num);
                    this.offset += 4;
                    return;
                }
            }
        }
        this.ensureCapacity(9);
        this.uint8[this.offset++] = 0xcb;
        this.view.setFloat64(this.offset, num);
        this.offset += 8;
    };
    Encoder.prototype.encodeStringHeader = function (length) {
        if (length <= 31)
            this.u8(160 | length);
        else if (length <= 0xff) {
            this.u8(0xd9);
            this.u8(length);
        }
        else if (length <= 0xffff) {
            this.u8(0xda);
            this.u16(length);
        }
        else if (length <= 0xffffffff) {
            this.u8(0xdb);
            this.u32(length);
        }
    };
    Encoder.prototype.encodeString = function (str) {
        var length = str.length;
        var maxSize = length * 4;
        this.ensureCapacity(5 + maxSize);
        var uint8 = this.uint8;
        var lengthOffset = this.offset;
        if (maxSize <= 31)
            this.offset++;
        else if (maxSize <= 0xff) {
            uint8[this.offset++] = 0xd9;
            lengthOffset = this.offset;
            this.offset++;
        }
        else if (maxSize <= 0xffff) {
            uint8[this.offset++] = 0xda;
            lengthOffset = this.offset;
            this.offset += 2;
        }
        else {
            uint8[this.offset++] = 0xdb;
            lengthOffset = this.offset;
            this.offset += 4;
        }
        var offset = this.offset;
        var pos = 0;
        while (pos < length) {
            var value = str.charCodeAt(pos++);
            if ((value & 0xffffff80) === 0) {
                uint8[offset++] = value;
                continue;
            }
            else if ((value & 0xfffff800) === 0) {
                uint8[offset++] = ((value >> 6) & 0x1f) | 0xc0;
            }
            else {
                if (value >= 0xd800 && value <= 0xdbff) {
                    if (pos < length) {
                        var extra = str.charCodeAt(pos);
                        if ((extra & 0xfc00) === 0xdc00) {
                            pos++;
                            value = ((value & 0x3ff) << 10) + (extra & 0x3ff) + 0x10000;
                        }
                    }
                }
                if ((value & 0xffff0000) === 0) {
                    uint8[offset++] = ((value >> 12) & 0x0f) | 0xe0;
                    uint8[offset++] = ((value >> 6) & 0x3f) | 0x80;
                }
                else {
                    uint8[offset++] = ((value >> 18) & 0x07) | 0xf0;
                    uint8[offset++] = ((value >> 12) & 0x3f) | 0x80;
                    uint8[offset++] = ((value >> 6) & 0x3f) | 0x80;
                }
            }
            uint8[offset++] = (value & 0x3f) | 0x80;
        }
        this.offset = offset;
        if (maxSize <= 31)
            uint8[lengthOffset] = 160 | (offset - lengthOffset - 1);
        else if (maxSize <= 0xff)
            uint8[lengthOffset] = offset - lengthOffset - 1;
        else if (maxSize <= 0xffff)
            this.view.setUint16(lengthOffset, offset - lengthOffset - 2);
        else
            this.view.setUint32(lengthOffset, offset - lengthOffset - 4);
    };
    Encoder.prototype.encodeArrayHeader = function (length) {
        if (length <= 15)
            this.u8(144 | length);
        else if (length <= 0xffff) {
            this.u8(0xdc);
            this.u16(length);
        }
        else if (length <= 0xffffffff) {
            this.u8(0xdd);
            this.u32(length);
        }
    };
    Encoder.prototype.encodeArray = function (arr) {
        var length = arr.length;
        if (length <= 15)
            this.u8(144 | length);
        else if (length <= 0xffff) {
            this.u8(0xdc);
            this.u16(length);
        }
        else if (length <= 0xffffffff) {
            this.u8(0xdd);
            this.u32(length);
        }
        else
            return;
        for (var i = 0; i < length; i++)
            this.encodeAny(arr[i]);
    };
    Encoder.prototype.encodeObjectHeader = function (length) {
        if (length <= 15)
            this.u8(128 | length);
        else if (length <= 0xffff) {
            this.u8(0xde);
            this.u16(length);
        }
        else if (length <= 0xffffffff) {
            this.u8(0xdf);
            this.u32(length);
        }
    };
    Encoder.prototype.encodeObject = function (obj) {
        var keys = Object.keys(obj);
        var length = keys.length;
        if (length <= 15)
            this.u8(128 | length);
        else if (length <= 0xffff) {
            this.u8(0xde);
            this.u16(length);
        }
        else if (length <= 0xffffffff) {
            this.u8(0xdf);
            this.u32(length);
        }
        else
            return;
        for (var i = 0; i < length; i++) {
            var key = keys[i];
            this.encodeString(key);
            this.encodeAny(obj[key]);
        }
    };
    Encoder.prototype.buf = function (buf, length) {
        this.ensureCapacity(length);
        this.uint8.set(buf, this.offset);
        this.offset += length;
    };
    Encoder.prototype.encodeExtHeader = function (type, length) {
        switch (length) {
            case 1:
                this.u16((0xd4 << 8) | type);
                break;
            case 2:
                this.u16((0xd5 << 8) | type);
                break;
            case 4:
                this.u16((0xd6 << 8) | type);
                break;
            case 8:
                this.u16((0xd7 << 8) | type);
                break;
            case 16:
                this.u16((0xd8 << 8) | type);
                break;
            default:
                if (length <= 0xff) {
                    this.u16((0xc7 << 8) | length);
                    this.u8(type);
                }
                else if (length <= 0xffff) {
                    this.u8(0xc8);
                    this.u16(length);
                    this.u8(type);
                }
                else if (length <= 0xffffffff) {
                    this.u8(0xc9);
                    this.u32(length);
                    this.u8(type);
                }
        }
    };
    Encoder.prototype.encodeExt = function (ext) {
        var type = ext.type, buf = ext.buf;
        var length = buf.byteLength;
        this.encodeExtHeader(type, length);
        this.buf(buf, length);
    };
    return Encoder;
}(encoder_1.Encoder));
exports.Encoder = Encoder;
